# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import binascii
import json
import logging
import os
from dataclasses import dataclass
from hashlib import blake2b
from pathlib import Path
from queue import Queue
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import xxhash
from scalecodec import GenericCall, GenericExtrinsic, ScaleBytes, ScaleType, ss58_decode
from scalecodec.base import RuntimeConfigurationObject, ScaleDecoder
from scalecodec.type_registry import load_type_registry_preset
from urllib3.util import parse_url
from websocket import WebSocket, create_connection

from tikka.libs.keypair import Keypair


@dataclass
class ExtrinsicReceipt:
    """
    ExtrinsicReceipt class
    """

    extrinsic_hash: str
    block_hash: Optional[str] = None
    finalized: bool = False
    is_success: bool = False


class SubstrateClient:
    """
    SubstrateClient class
    """

    METADATA_FILE_BASENAME = "substrate_client_metadata_"
    CONNECTIONS_POOL_SIZE = 5

    def __init__(
        self,
        url: str,
        websocket_options: Optional[dict] = None,
        metadata_dir: str = "~/.substrate_client",
    ):
        """
        Init SubstrateClient instance

        :param url:
        """
        # static variables
        self.url: str = url
        self.metadata_dir = metadata_dir
        self.connections_pool: Queue = Queue(self.CONNECTIONS_POOL_SIZE)
        # Websocket connection options
        self.websocket_options = websocket_options or {}

        # variables per connection
        self._request_id: Dict[int, int] = {}
        self._rpc_message_queue: Dict[int, list] = {}

        supported_schemes = ("ws", "wss")
        url_scheme = parse_url(self.url).scheme
        if url_scheme not in supported_schemes:
            raise Exception(f"Scheme {url_scheme} not supported")

        if not os.path.exists(Path(metadata_dir).expanduser()):
            os.mkdir(Path(metadata_dir).expanduser())

        # create pool of connections
        logging.debug("Connecting to {} ...".format(self.url))
        for i in range(self.CONNECTIONS_POOL_SIZE):
            connection: WebSocket = create_connection(
                self.url, **self.websocket_options
            )
            self.connections_pool.put(connection)
            self._request_id[connection.fileno()] = 0
            self._rpc_message_queue[connection.fileno()] = []

        runtime_info = self.get_runtime_info()
        if runtime_info is None:
            raise Exception("No runtime info found")
        self.transaction_version = runtime_info["transactionVersion"]
        self.runtime_version = runtime_info["specVersion"]
        logging.debug(f"runtime version: {self.runtime_version}")
        logging.debug(f"transaction version: {self.transaction_version}")

        self.runtime_config: RuntimeConfigurationObject = RuntimeConfigurationObject()

        # This types are all hardcoded types needed to decode metadata types
        self.runtime_config.update_type_registry(load_type_registry_preset(name="core"))

        self.metadata: ScaleType = self.get_metadata()

        # Check if PortableRegistry is present in metadata (V14+), otherwise fall back on legacy type registry (<V14)
        if self.implements_scaleinfo():
            self.runtime_config.add_portable_registry(self.metadata)

        # Set active runtime version
        self.runtime_config.set_active_spec_version_id(self.runtime_version)
        self.runtime_config.ss58_format = self.get_constant("System", "SS58Prefix")

        # Set runtime compatibility flags
        try:
            _ = self.runtime_config.create_scale_object("sp_weights::weight_v2::Weight")
            self.runtime_config.update_type_registry_types(
                {"Weight": "sp_weights::weight_v2::Weight"}
            )
        except NotImplementedError:
            self.runtime_config.update_type_registry_types({"Weight": "WeightV1"})

    def close(self):
        """
        Close websocket connections in pool

        :return:
        """
        while not self.connections_pool.not_empty:
            connection: WebSocket = self.connections_pool.get()
            connection.close()
            connection_id = connection.fileno()
            del self._request_id[connection_id]
            del self._rpc_message_queue[connection_id]

        logging.debug("SubstrateClient connection closed.")

    def get_constant(self, module_name, constant_name) -> Optional[ScaleType]:
        """
        Returns the decoded `ScaleType` object of the constant for given module name, call function name

        :param module_name: Module name
        :param constant_name: Parameter name

        :return:
        """
        constant = None
        for module_idx, module in enumerate(self.metadata.pallets):

            if module_name == module.name and module.constants:

                for _constant in module.constants:
                    if constant_name == _constant.value["name"]:
                        constant = _constant

                if constant:

                    obj = self.runtime_config.create_scale_object(
                        type_string=constant.type,
                        data=ScaleBytes(constant.constant_value),
                        metadata=self.metadata,
                    )

                    obj.decode(check_remaining=True)

                    constant = obj.value

        return constant

    def rpc_request(
        self,
        method: str,
        params: Optional[list] = None,
        result_handler: Optional[Callable] = None,
    ) -> dict:
        """
        Send JSON-RPC request to server

        :param method: Method name
        :param params: list of parameters
        :param result_handler: Function to call to handle result

        :return:
        """
        # get a connection from pool
        connection: WebSocket = self.connections_pool.get()
        connection_id = connection.fileno()

        self._request_id[connection_id] += 1
        request_id = self._request_id[connection_id]

        params = params or []
        request = {
            "jsonrpc": "2.0",
            "method": method,
            "params": params,
            "id": request_id,
        }

        logging.debug('RPC request #{}: "{}"'.format(request_id, method))

        connection.send(json.dumps(request))

        update_nr = 0
        json_body = None
        subscription_id = None

        while json_body is None:
            # Search for subscriptions
            for message, remove_message in list_remove_iter(
                self._rpc_message_queue[connection_id]
            ):

                # Check if result message is matching request ID
                if "id" in message and message["id"] == request_id:

                    remove_message()

                    # Check if response has error
                    if "error" in message:
                        self.connections_pool.put(
                            connection
                        )  # Return connection back to the pool
                        raise Exception(message["error"])

                    # If result handler is set, pass result through and loop until handler return value is set
                    if callable(result_handler):

                        # Set subscription ID and only listen to messages containing this ID
                        subscription_id = message["result"]
                        logging.debug(
                            f"Websocket subscription [{subscription_id}] created"
                        )

                    else:
                        json_body = message

            # Process subscription updates
            for message, remove_message in list_remove_iter(
                self._rpc_message_queue[connection_id]
            ):
                # Check if message is meant for this subscription
                if (
                    callable(result_handler)
                    and "params" in message
                    and message["params"]["subscription"] == subscription_id
                ):

                    remove_message()

                    logging.debug(
                        f"Websocket result [{subscription_id} #{update_nr}]: {message}"
                    )

                    # Call result_handler with message for processing
                    callback_result = result_handler(
                        message, update_nr, subscription_id
                    )
                    if callback_result is not None:
                        json_body = callback_result

                    update_nr += 1

            # Read one more message to queue
            if json_body is None:
                self._rpc_message_queue[connection_id].append(
                    json.loads(connection.recv())
                )

        self.connections_pool.put(connection)  # Return connection back to the pool

        return json_body

    def get_runtime_info(self) -> Optional[Any]:
        """
        Return runtime info from server for metadata caching

        :return:
        """
        response = self.rpc_request("state_getRuntimeVersion")
        if "error" in response:
            raise Exception(response["error"]["message"])

        return response.get("result")

    def get_metadata(self) -> ScaleType:
        """
        Return metadata json from server

        :return:
        """
        metadata_filepath = (
            Path(self.metadata_dir)
            .joinpath(self.METADATA_FILE_BASENAME + str(self.runtime_version) + ".json")
            .expanduser()
        )

        if not os.path.exists(metadata_filepath):
            response = self.rpc_request("state_getMetadata")
            metadata_scale_encoded = response.get("result")
            with open(metadata_filepath, "w") as file:
                json.dump(metadata_scale_encoded, file)
            logging.debug(f"metadata saved to {metadata_filepath}")
        else:
            with open(metadata_filepath, "r") as file:
                metadata_scale_encoded = json.load(file)
            logging.debug(f"metadata loaded from {metadata_filepath}")

        # Decode retrieved metadata from the RPC
        metadata = self.runtime_config.create_scale_object(
            "MetadataVersioned", data=ScaleBytes(metadata_scale_encoded)
        )
        metadata.decode()

        return metadata

    def query(
        self, module: str, storage_function: str, params: Optional[list] = None
    ) -> Any:
        """
        Return a ScaleType instance of the response of the query

        :param module: Module name
        :param storage_function: Storage function name
        :param params: List of parameters

        :return:
        """
        if self.metadata is None:
            raise Exception("Metadata empty")

        # Search storage call in metadata
        metadata_pallet = self.metadata.get_metadata_pallet(module)

        if not metadata_pallet:
            raise Exception(f'Pallet "{module}" not found')

        storage_item = metadata_pallet.get_storage_function(storage_function)

        if not metadata_pallet or not storage_item:
            raise Exception(f'Storage function "{module}.{storage_function}" not found')

        params = params or []
        _scale_type = self.get_scale_type(module, storage_function, params)
        _storage_key = _scale_type.storage_hash

        # SCALE type string of value
        param_types = storage_item.get_params_type_string()
        # value_scale_type = storage_item.get_value_type_string()

        if len(params) != len(param_types):
            raise ValueError(
                f"Storage function requires {len(param_types)} parameters, {len(params)} given"
            )

        response = self.rpc_request("state_getStorage", [_storage_key])

        if "error" in response:
            raise Exception(response["error"]["message"])

        scale_data = response["result"]
        if scale_data is None:
            return None
        #
        # if 'error' in response:
        #     raise Exception(response['error']['message'])
        #
        # if 'result' in response:
        #     if value_scale_type:
        #
        #         if response.get('result') is not None:
        #             query_value = response.get('result')
        #         elif storage_item.value['modifier'] == 'Default':
        #             # Fallback to default value of storage function if no result
        #             query_value = storage_item.value_object['default'].value_object
        #         else:
        #             # No result is interpreted as an Option<...> result
        #             value_scale_type = f'Option<{value_scale_type}>'
        #             query_value = storage_item.value_object['default'].value_object
        #
        #         obj = self.runtime_config.create_scale_object(
        #             type_string=value_scale_type,
        #             data=ScaleBytes(query_value),
        #             metadata=self.metadata
        #         )
        #         obj.decode(check_remaining=True)
        #         obj.meta_info = {'result_found': response.get('result') is not None}
        #
        #         return obj

        return _scale_type.decode(ScaleBytes(scale_data))

    def query_multi(
        self,
        storage_functions: List[Tuple[str, str, list]],
    ) -> list:
        """
        Query multiple storage keys in one request.

        ```
        storage_functions = [
            "System", "Account", ["F4xQKRUagnSGjFqafyhajLs94e7Vvzvr8ebwYJceKpr8R7T"],
            "System", "Account", ["GSEX8kR4Kz5UZGhvRUCJG93D5hhTAoVZ5tAe6Zne7V42DSi"]
        ]

        result = substrate.query_multi(storage_keys)
        ```

        :param storage_functions: List of Storage function as tuples (module, function, [params,...])

        :return:
        """
        if self.metadata is None:
            raise Exception("Metadata empty")

        storage_keys = []
        scale_types: Dict[str, ScaleType] = {}
        for module, storage_function, params in storage_functions:
            # Search storage call in metadata
            metadata_pallet = self.metadata.get_metadata_pallet(module)

            if not metadata_pallet:
                raise Exception(f'Pallet "{module}" not found')

            storage_item = metadata_pallet.get_storage_function(storage_function)

            if not metadata_pallet or not storage_item:
                raise Exception(
                    f'Storage function "{module}.{storage_function}" not found'
                )

            _scale_type = self.get_scale_type(module, storage_function, params)
            _storage_key = _scale_type.storage_hash
            storage_keys.append(_storage_key)
            scale_types[_storage_key] = _scale_type

        # Retrieve corresponding value
        response = self.rpc_request("state_queryStorageAt", [storage_keys])

        if "error" in response:
            raise Exception(response["error"]["message"])
        result = []

        storage_key_map = {"0x" + s: s for s in storage_keys}
        for result_group in response["result"]:
            for change_storage_key, change_data in result_group["changes"]:
                # Decode result for specified storage_key
                storage_key = storage_key_map[change_storage_key]
                if change_data is not None:
                    change_data = ScaleBytes(change_data)
                    result.append(scale_types[storage_key].decode(change_data))
                else:
                    result.append(None)

        return result

    def compose_call(
        self, call_module: str, call_function: str, call_params: Optional[dict] = None
    ) -> GenericCall:
        """
        Composes a call payload which can be used in an extrinsic.

        :param call_module: Name of the runtime module e.g. Balances
        :param call_function: Name of the call function e.g. transfer
        :param call_params: This is a dict containing the params of the call.
                             e.g. `{'dest': 'EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk', 'value': 1000000000000}`

        :return:
        """

        if call_params is None:
            call_params = {}

        call = self.runtime_config.create_scale_object(
            type_string="Call", metadata=self.metadata
        )

        call.encode(
            {
                "call_module": call_module,
                "call_function": call_function,
                "call_args": call_params,
            }
        )

        return call

    def submit_extrinsic(
        self,
        extrinsic: GenericExtrinsic,
        wait_for_inclusion: bool = False,
        wait_for_finalization: bool = False,
    ) -> "ExtrinsicReceipt":
        """
        Submit an extrinsic to the connected node, with the possibility to wait until the extrinsic is included
         in a block and/or the block is finalized. The receipt returned provided information about the block and
         triggered events

        :param extrinsic: Extrinsic The extrinsic to be sent to the network
        :param wait_for_inclusion: wait until extrinsic is included in a block (only works for websocket connections)
        :param wait_for_finalization: wait until extrinsic is finalized (only works for websocket connections)

        :return:
        """

        # Check requirements
        if not isinstance(extrinsic, GenericExtrinsic):
            raise TypeError("'extrinsic' must be of type Extrinsics")

        def result_handler(message: dict, update_nr: int, subscription_id: int):
            """
            Parse subscription message filtered by subscription_id
            and unsubscribe when included or finalized

            :param message: Message dict
            :param update_nr: Update number
            :param subscription_id: Subscription ID number

            :return:
            """
            # Check if extrinsic is included and finalized
            if "params" in message and type(message["params"]["result"]) is dict:

                # Convert result enum to lower for backwards compatibility
                message_result = {
                    k.lower(): v for k, v in message["params"]["result"].items()
                }

                if "finalized" in message_result and wait_for_finalization:
                    self.rpc_request("author_unwatchExtrinsic", [subscription_id])
                    return {
                        "block_hash": message_result["finalized"],
                        "extrinsic_hash": "0x{}".format(extrinsic.extrinsic_hash.hex()),
                        "finalized": True,
                    }
                elif (
                    "inblock" in message_result
                    and wait_for_inclusion
                    and not wait_for_finalization
                ):
                    self.rpc_request("author_unwatchExtrinsic", [subscription_id])
                    return {
                        "block_hash": message_result["inblock"],
                        "extrinsic_hash": "0x{}".format(extrinsic.extrinsic_hash.hex()),
                        "finalized": False,
                    }

        if wait_for_inclusion or wait_for_finalization:
            response = self.rpc_request(
                "author_submitAndWatchExtrinsic",
                [str(extrinsic.data)],
                result_handler=result_handler,
            )

            result = ExtrinsicReceipt(
                extrinsic_hash=response["extrinsic_hash"],
                block_hash=response["block_hash"],
                finalized=response["finalized"],
                is_success=True,
            )
        else:
            response = self.rpc_request("author_submitExtrinsic", [str(extrinsic.data)])

            if "result" not in response:
                raise Exception(response.get("error"))

            result = ExtrinsicReceipt(
                extrinsic_hash=response["result"], is_success=True
            )

        return result

    def get_payment_info(self, call: GenericCall, keypair: Keypair):
        """
        Retrieves fee estimation via RPC for given extrinsic

        Return Dict with payment info:

        {'class': 'normal', 'partialFee': 151000000, 'weight': {'ref_time': 143322000}}

        :param call: Call object to estimate fees for
        :param keypair: Keypair of the sender,
                        does not have to include private key because no valid signature is required

        :return:
        """

        # Check requirements
        if not isinstance(call, GenericCall):
            raise TypeError("'call' must be of type Call")

        if not isinstance(keypair, Keypair):
            raise TypeError("'keypair' must be of type Keypair")

        # No valid signature is required for fee estimation
        signature = "0x" + "00" * 64

        # Create extrinsic
        extrinsic = self.create_signed_extrinsic(
            call=call, keypair=keypair, signature=signature
        )

        # if self.supports_rpc_method('state_call'):
        #     extrinsic_len = self.runtime_config.create_scale_object('u32')
        #     extrinsic_len.encode(len(extrinsic.data))
        #
        #     result = self.runtime_call("TransactionPaymentApi", "query_info", [extrinsic, extrinsic_len])
        #
        #     return result.value
        # else:
        # Backwards compatibility; deprecated RPC method
        payment_info = self.rpc_request("payment_queryInfo", [str(extrinsic.data)])

        # convert partialFee to int
        if "result" in payment_info:
            payment_info["result"]["partialFee"] = int(
                payment_info["result"]["partialFee"]
            )

            if type(payment_info["result"]["weight"]) is int:
                # Transform format to WeightV2 if applicable as per https://github.com/paritytech/substrate/pull/12633
                try:
                    weight_obj = self.runtime_config.create_scale_object(
                        "sp_weights::weight_v2::Weight"
                    )
                    if weight_obj is not None:
                        payment_info["result"]["weight"] = {
                            "ref_time": payment_info["result"]["weight"],
                            "proof_size": 0,
                        }
                except NotImplementedError:
                    pass

            return payment_info["result"]
        else:
            raise Exception(payment_info["error"]["message"])

    ###########################################
    def get_chain_finalised_head(self) -> Optional[Any]:
        """
        A pass-though to existing JSONRPC method `chain_getFinalizedHead`

        :return:
        """
        response = self.rpc_request("chain_getFinalizedHead", [])

        if response is not None:
            if "error" in response:
                raise Exception(response["error"]["message"])

            return response.get("result")

    def get_block_number(self, block_hash: str) -> int:
        """
        A convenience method to get the block number for given block_hash

        :param block_hash: Hash of block

        :return:
        """
        response = self.rpc_request("chain_getHeader", [block_hash])

        if "error" in response:
            raise Exception(response["error"]["message"])

        elif "result" in response:

            if response["result"]:
                return int(response["result"]["number"], 16)

        raise Exception(f"get_block_number error: {response['result']}")

    def get_block_hash(self, block_id: Optional[int] = None) -> Optional[Any]:
        """
        A pass-though to existing JSONRPC method `chain_getBlockHash`

        :param block_id: Block number

        :return:
        """
        response = self.rpc_request("chain_getBlockHash", [block_id])

        if "error" in response:
            raise Exception(response["error"]["message"])
        else:
            return response.get("result")

    def generate_signature_payload(
        self,
        call: GenericCall,
        era: Optional[Union[str, dict]] = None,
        nonce: int = 0,
        tip: int = 0,
        tip_asset_id: Optional[int] = None,
        include_call_length: bool = False,
    ) -> ScaleBytes:
        """
        Generate signature payload for extrinsic call

        :param call: GenericCall instance
        :param era: Dict of Era
        :param nonce: Signer account Nonce
        :param tip: Tip amount
        :param tip_asset_id: Tip asset ID
        :param include_call_length: True to include call length

        :return:
        """
        if self.metadata is None:
            raise Exception("Metadata empty")

        # Retrieve genesis hash
        genesis_hash = self.get_block_hash(0)

        if not era:
            era = "00"

        if era == "00":
            # Immortal extrinsic
            block_hash = genesis_hash
        else:
            # Determine mortality of extrinsic
            era_obj = self.runtime_config.create_scale_object("Era")

            if isinstance(era, dict) and "current" not in era and "phase" not in era:
                raise ValueError(
                    'The era dict must contain either "current" or "phase" element to encode a valid era'
                )

            era_obj.encode(era)
            block_hash = self.get_block_hash(block_id=era_obj.birth(era.get("current")))  # type: ignore

        # Create signature payload
        signature_payload = self.runtime_config.create_scale_object(
            "ExtrinsicPayloadValue"
        )

        # Process signed extensions in metadata
        if "signed_extensions" in self.metadata[1][1]["extrinsic"]:

            # Base signature payload
            signature_payload.type_mapping = [["call", "CallBytes"]]

            # Add signed extensions to payload
            signed_extensions = self.metadata.get_signed_extensions()

            if "CheckMortality" in signed_extensions:
                signature_payload.type_mapping.append(
                    ["era", signed_extensions["CheckMortality"]["extrinsic"]]
                )

            if "CheckEra" in signed_extensions:
                signature_payload.type_mapping.append(
                    ["era", signed_extensions["CheckEra"]["extrinsic"]]
                )

            if "CheckNonce" in signed_extensions:
                signature_payload.type_mapping.append(
                    ["nonce", signed_extensions["CheckNonce"]["extrinsic"]]
                )

            if "ChargeTransactionPayment" in signed_extensions:
                signature_payload.type_mapping.append(
                    ["tip", signed_extensions["ChargeTransactionPayment"]["extrinsic"]]
                )

            if "ChargeAssetTxPayment" in signed_extensions:
                signature_payload.type_mapping.append(
                    ["asset_id", signed_extensions["ChargeAssetTxPayment"]["extrinsic"]]
                )

            if "CheckMetadataHash" in signed_extensions:
                signature_payload.type_mapping.append(
                    ["mode", signed_extensions["CheckMetadataHash"]["extrinsic"]]
                )

            if "CheckSpecVersion" in signed_extensions:
                signature_payload.type_mapping.append(
                    [
                        "spec_version",
                        signed_extensions["CheckSpecVersion"]["additional_signed"],
                    ]
                )

            if "CheckTxVersion" in signed_extensions:
                signature_payload.type_mapping.append(
                    [
                        "transaction_version",
                        signed_extensions["CheckTxVersion"]["additional_signed"],
                    ]
                )

            if "CheckGenesis" in signed_extensions:
                signature_payload.type_mapping.append(
                    [
                        "genesis_hash",
                        signed_extensions["CheckGenesis"]["additional_signed"],
                    ]
                )

            if "CheckMortality" in signed_extensions:
                signature_payload.type_mapping.append(
                    [
                        "block_hash",
                        signed_extensions["CheckMortality"]["additional_signed"],
                    ]
                )

            if "CheckEra" in signed_extensions:
                signature_payload.type_mapping.append(
                    ["block_hash", signed_extensions["CheckEra"]["additional_signed"]]
                )

            if "CheckMetadataHash" in signed_extensions:
                signature_payload.type_mapping.append(
                    [
                        "metadata_hash",
                        signed_extensions["CheckMetadataHash"]["additional_signed"],
                    ]
                )

        if include_call_length:

            length_obj = self.runtime_config.create_scale_object("Bytes")
            call_data = str(length_obj.encode(str(call.data)))

        else:
            call_data = str(call.data)

        payload_dict = {
            "call": call_data,
            "era": era,
            "nonce": nonce,
            "tip": tip,
            "spec_version": self.runtime_version,
            "genesis_hash": genesis_hash,
            "block_hash": block_hash,
            "transaction_version": self.transaction_version,
            "asset_id": {"tip": tip, "asset_id": tip_asset_id},
            "metadata_hash": None,
            "mode": "Disabled",
        }

        signature_payload.encode(payload_dict)

        if signature_payload.data.length > 256:
            return ScaleBytes(
                data=blake2b(signature_payload.data.data, digest_size=32).digest()
            )

        return signature_payload.data

    def create_signed_extrinsic(
        self,
        call: GenericCall,
        keypair: Keypair,
        era: Optional[Union[str, dict]] = None,
        nonce: Optional[int] = None,
        tip: int = 0,
        tip_asset_id: Optional[int] = None,
        signature: Optional[Union[bytes, str]] = None,
    ) -> GenericExtrinsic:
        """
        Creates an extrinsic signed by given account details

        :param call: GenericCall to create extrinsic for
        :param keypair: Keypair used to sign the extrinsic
        :param era: Specify mortality in blocks in follow format: {'period': [amount_blocks]}
             If omitted the extrinsic is immortal
        :param nonce: nonce to include in extrinsics, if omitted the current nonce is retrieved on-chain
        :param tip: The tip for the block author to gain priority during network congestion
        :param tip_asset_id: Optional asset ID with which to pay the tip
        :param signature: Optionally provide signature if externally signed

        :return:
        """
        if self.metadata is None:
            raise Exception("Metadata empty")

        # Check requirements
        if not isinstance(call, GenericCall):
            raise TypeError("'call' must be of type Call")

        # Check if extrinsic version is supported
        if self.metadata[1][1]["extrinsic"]["version"].value != 4:
            raise NotImplementedError(
                f"Extrinsic version {self.metadata[1][1]['extrinsic']['version']} not supported"
            )

        # Retrieve nonce
        if nonce is None:
            response = self.rpc_request(
                "system_accountNextIndex", [keypair.ss58_address]
            )
            nonce = response.get("result") or 0

        # Process era
        if era is None:
            era = "00"
        else:
            if isinstance(era, dict) and "current" not in era and "phase" not in era:
                # Retrieve current block id
                era["current"] = self.get_block_number(self.get_chain_finalised_head())  # type: ignore

        if signature is not None:

            if type(signature) is str and signature[0:2] == "0x":
                signature = bytes.fromhex(signature[2:])  # type: ignore

            # Check if signature is a MultiSignature and contains signature version
            if len(signature) == 65:
                signature_version = signature[0]
                signature = signature[1:]
            else:
                signature_version = keypair.crypto_type

        else:
            # Create signature payload
            signature_payload = self.generate_signature_payload(
                call=call, era=era, nonce=nonce, tip=tip, tip_asset_id=tip_asset_id
            )

            # Set Signature version to crypto type of keypair
            signature_version = keypair.crypto_type

            # Sign payload
            signature = keypair.sign(signature_payload)

        # Create extrinsic
        extrinsic = self.runtime_config.create_scale_object(
            type_string="Extrinsic", metadata=self.metadata
        )

        value = {
            "account_id": f"0x{keypair.public_key.hex()}",
            "signature": f"0x{signature.hex()}",  # type: ignore
            "call_function": call.value["call_function"],
            "call_module": call.value["call_module"],
            "call_args": call.value["call_args"],
            "nonce": nonce,
            "era": era,
            "tip": tip,
            "asset_id": {"tip": tip, "asset_id": tip_asset_id},
            "mode": "Disabled",
        }

        # Check if ExtrinsicSignature is MultiSignature, otherwise omit signature_version
        signature_cls = self.runtime_config.get_decoder_class("ExtrinsicSignature")
        if issubclass(signature_cls, self.runtime_config.get_decoder_class("Enum")):
            value["signature_version"] = signature_version

        extrinsic.encode(value)

        return extrinsic

    #############################################################
    def implements_scaleinfo(self) -> Optional[bool]:
        """
        Returns True if current runtime implementation a `PortableRegistry` (`MetadataV14` and higher)

        Returns
        -------
        bool
        """
        if self.metadata:
            return self.metadata.portable_registry is not None
        return None

    def get_scale_type(
        self, module: str, module_function: str, params: list
    ) -> ScaleType:
        """
        Return scale type

        :param module: Module name
        :param module_function: Module function name
        :param params: List of parameters

        :return:
        """
        if self.metadata is None:
            raise Exception("Metadata empty")

        # # Add the embedded type registry to the runtime config
        # self.runtime_config.add_portable_registry(self.metadata)

        metadata_pallet = self.metadata.get_metadata_pallet(module)
        metadata_storage_function = metadata_pallet.get_storage_function(
            module_function
        )

        # Process specific type of storage function
        value_scale_type = metadata_storage_function.get_value_type_string()
        param_types = metadata_storage_function.get_params_type_string()
        storage_hash = xxh128(
            metadata_pallet.value["storage"]["prefix"].encode()
        ) + xxh128(module_function.encode())

        hashers = metadata_storage_function.get_param_hashers()

        hasher_functions = {
            "Blake2_256": blake2_256,
            "Blake2_128": blake2_128,
            "Blake2_128Concat": blake2_128_concat,
            "Twox128": xxh128,
            "Twox64Concat": two_x64_concat,
            "Identity": lambda x: x,
        }

        # Encode parameters
        params_encoded = []
        if params:
            for idx, param in enumerate(params):
                if type(param) is ScaleBytes:
                    # Already encoded
                    params_encoded.append(param)
                else:
                    param = convert_storage_parameter(
                        self.runtime_config, param_types[idx], param
                    )
                    param_obj = self.runtime_config.create_scale_object(
                        type_string=param_types[idx]
                    )
                    params_encoded.append(param_obj.encode(param))

            for idx, param in enumerate(params_encoded):
                # Get hasher assiociated with param
                try:
                    param_hasher = hashers[idx]
                except IndexError:
                    raise ValueError(f"No hasher found for param #{idx + 1}")

                params_key = bytes()

                # Convert param to bytes
                if type(param) is str:
                    params_key += binascii.unhexlify(param)
                elif type(param) is ScaleBytes:
                    params_key += param.data
                elif isinstance(param, ScaleDecoder):
                    params_key += param.data.data

                if not param_hasher:
                    param_hasher = "Twox128"

                if param_hasher in hasher_functions:
                    storage_hash += hasher_functions[param_hasher](params_key)
                else:
                    raise ValueError('Unknown storage hasher "{}"'.format(param_hasher))

        scale_object = self.runtime_config.create_scale_object(
            value_scale_type, metadata=self.metadata
        )
        scale_object.storage_hash = storage_hash.hex()
        return scale_object


def blake2_256(data):
    """
    Helper function to calculate a 32 bytes Blake2b hash for provided data, used as key for Substrate storage items

    :param data: Data to encode

    :return:
    """
    return blake2b(data, digest_size=32).digest()


def blake2_128(data):
    """
    Helper function to calculate a 16 bytes Blake2b hash for provided data, used as key for Substrate storage items

    :param data: Data to encode

    :return:
    """
    return blake2b(data, digest_size=16).digest()


def blake2_128_concat(data):
    """
    Helper function to calculate a 16 bytes Blake2b hash for provided data, concatenated with data, used as key
    for Substrate storage items

    :param data: Data to encode

    :return:
    """
    return blake2b(data, digest_size=16).digest() + data


def two_x64_concat(data):
    """
    Helper function to calculate a xxh64 hash with concatenated data for provided data,
    used as key for several Substrate

    :param data: Data to encode

    :return:
    """
    storage_key = bytearray(xxhash.xxh64(data, seed=0).digest())
    storage_key.reverse()

    return storage_key + data


def xxh128(data):
    """
    Helper function to calculate a 2 concatenated xxh64 hash for provided data, used as key for several Substrate

    :param data: Data to encode

    :return:
    """
    storage_key1 = bytearray(xxhash.xxh64(data, seed=0).digest())
    storage_key1.reverse()

    storage_key2 = bytearray(xxhash.xxh64(data, seed=1).digest())
    storage_key2.reverse()

    return storage_key1 + storage_key2


def convert_storage_parameter(
    runtime_config: RuntimeConfigurationObject, scale_type: str, value: Any
):
    """


    :param runtime_config: RuntimeConfigurationObject instance
    :param scale_type: Scale type string
    :param value: Parameter value

    :return:
    """
    if type(value) is bytes:
        value = f"0x{value.hex()}"

    if scale_type == "AccountId":
        if value[0:2] != "0x":
            return "0x{}".format(ss58_decode(value, runtime_config.ss58_format))

    return value


def list_remove_iter(xs: list):
    """
    List remover

    :param xs: List to handle

    :return:
    """
    removed = False

    def remove():
        nonlocal removed
        removed = True

    i = 0
    while i < len(xs):
        removed = False
        yield xs[i], remove
        if removed:
            xs.pop(i)
        else:
            i += 1


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)

    _client = SubstrateClient("ws://localhost:9944")

    alice_address = "5GrwvaEF5zXb26Fz9rcQpDWS57CtERHpNehXCPcNoHGKutQY"
    bob_address = "5FHneW46xGXgs5mUiveU4sbTyGBzmstUspZC92UhjJM694ty"
    alice_keypair = Keypair.create_from_uri(
        "bottom drive obey lake curtain smoke basket hold race lonely fit walk//Alice"
    )

    # response = client.query("System", "Account", [alice_address])
    # print(response)

    # _response = _client.query_multi(
    #     [("System", "Account", [alice_address]), ("System", "Account", [bob_address])]
    # )
    # print(_response)
    #
    # _params = {"" "dest": bob_address, "value": 1100}
    #
    # _call = _client.compose_call(
    #     call_module="Balances",
    #     call_function="transfer_keep_alive",
    #     call_params=_params,
    # )
    #
    # _extrinsic = _client.create_signed_extrinsic(call=_call, keypair=alice_keypair)
    #
    # _receipt = _client.submit_extrinsic(
    #     _extrinsic,
    #     wait_for_inclusion=True,
    #     wait_for_finalization=False,
    # )
    # logging.debug(
    #     "Extrinsic '%s' sent and included in block '%s'",
    #     _receipt.extrinsic_hash,
    #     _receipt.block_hash,
    # )
    #
    # _response = _client.query_multi(
    #     [("System", "Account", [alice_address]), ("System", "Account", [bob_address])]
    # )
    # print(_response)

    result = _client.query("Identity", "Identities", [1])
    print(result)

    _client.close()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'tikka/slots/pyqt/resources/gui/windows/category_add.ui'
#
# Created by: PyQt5 UI code generator 5.15.10
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_addCategoryDialog(object):
    def setupUi(self, addCategoryDialog):
        addCategoryDialog.setObjectName("addCategoryDialog")
        addCategoryDialog.resize(400, 111)
        self.gridLayout = QtWidgets.QGridLayout(addCategoryDialog)
        self.gridLayout.setObjectName("gridLayout")
        self.buttonBox = QtWidgets.QDialogButtonBox(addCategoryDialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.gridLayout.addWidget(self.buttonBox, 2, 0, 1, 2, QtCore.Qt.AlignHCenter)
        self.nameLabel = QtWidgets.QLabel(addCategoryDialog)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.nameLabel.setFont(font)
        self.nameLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.nameLabel.setObjectName("nameLabel")
        self.gridLayout.addWidget(self.nameLabel, 0, 0, 1, 1)
        self.nameValueLineEdit = QtWidgets.QLineEdit(addCategoryDialog)
        self.nameValueLineEdit.setObjectName("nameValueLineEdit")
        self.gridLayout.addWidget(self.nameValueLineEdit, 0, 1, 1, 1)
        self.parentComboBox = QtWidgets.QComboBox(addCategoryDialog)
        self.parentComboBox.setObjectName("parentComboBox")
        self.gridLayout.addWidget(self.parentComboBox, 1, 1, 1, 1)
        self.parentLabel = QtWidgets.QLabel(addCategoryDialog)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.parentLabel.setFont(font)
        self.parentLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.parentLabel.setObjectName("parentLabel")
        self.gridLayout.addWidget(self.parentLabel, 1, 0, 1, 1)

        self.retranslateUi(addCategoryDialog)
        self.buttonBox.accepted.connect(addCategoryDialog.accept) # type: ignore
        self.buttonBox.rejected.connect(addCategoryDialog.reject) # type: ignore
        QtCore.QMetaObject.connectSlotsByName(addCategoryDialog)

    def retranslateUi(self, addCategoryDialog):
        #  _translate = QtCore.QCoreApplication.translate
        addCategoryDialog.setWindowTitle(_("Add category"))
        self.nameLabel.setText(_("Name"))
        self.parentLabel.setText(_("Parent"))



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
from datetime import datetime, timedelta
from typing import List, Optional

from tikka.domains.entities.smith import Smith, SmithStatus
from tikka.interfaces.adapters.network.connection import NetworkConnectionError
from tikka.interfaces.adapters.network.node.smiths import (
    NodeSmithsException,
    NodeSmithsInterface,
)
from tikka.libs.keypair import Keypair


class NodeSmiths(NodeSmithsInterface):
    """
    NodeSmiths class
    """

    status_map = {
        "Invited": SmithStatus.INVITED,
        "Pending": SmithStatus.PENDING,
        "Smith": SmithStatus.SMITH,
        "Excluded": SmithStatus.EXCLUDED,
    }

    def get_smith(self, identity_index: int) -> Optional[Smith]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NodeSmithsInterface.get_smith.__doc__
        )
        if (
            not self.node.connection.is_connected()
            or self.node.connection.client is None
        ):
            raise NodeSmithsException(NetworkConnectionError())

        try:
            result = self.node.connection.client.query(
                "SmithMembers", "Smiths", [identity_index]
            )
        except Exception as exception:
            logging.exception(exception)
            raise NodeSmithsException(exception)

        smith = None
        if result is not None:
            if result["expires_on"] is not None:
                expire_on_datetime = self.get_datetime_from_epoch(result["expires_on"])
            else:
                expire_on_datetime = None
            smith = Smith(
                identity_index=identity_index,
                status=self.status_map[result["status"]],
                expire_on=expire_on_datetime,
                certifications_received=result["received_certs"],
                certifications_issued=result["issued_certs"],
            )

        return smith

    def get_smiths(self, identity_indice: List[int]) -> List[Optional[Smith]]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NodeSmithsInterface.get_smiths.__doc__
        )
        if (
            not self.node.connection.is_connected()
            or self.node.connection.client is None
        ):
            raise NodeSmithsException(NetworkConnectionError())

        storage_functions = []
        for identity_index in identity_indice:
            storage_functions.append(("SmithMembers", "Smiths", [identity_index]))

        try:
            multi_result = self.node.connection.client.query_multi(storage_functions)
        except Exception as exception:
            logging.exception(exception)
            raise NodeSmithsException(exception)

        smiths: List[Optional[Smith]] = []
        for index, value_obj in enumerate(multi_result):
            if value_obj is not None:
                if value_obj["expires_on"] is not None:
                    expire_on_datetime = self.get_datetime_from_epoch(
                        value_obj["expires_on"]
                    )
                else:
                    expire_on_datetime = None

                smiths.append(
                    Smith(
                        identity_index=identity_indice[index],
                        status=self.status_map[value_obj["status"]],
                        expire_on=expire_on_datetime,
                        certifications_received=value_obj["received_certs"],
                        certifications_issued=value_obj["issued_certs"],
                    )
                )
            else:
                smiths.append(None)

        return smiths

    def invite(self, keypair: Keypair, identity_index: int) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NodeSmithsInterface.invite.__doc__
        )
        if (
            not self.node.connection.is_connected()
            or self.node.connection.client is None
        ):
            raise NodeSmithsException(NetworkConnectionError())

        params = {
            "receiver": identity_index,
        }

        try:
            call = self.node.connection.client.compose_call(
                call_module="SmithMembers",
                call_function="invite_smith",
                call_params=params,
            )
        except Exception as exception:
            logging.exception(exception)
            raise NodeSmithsException(exception)

        try:
            extrinsic = self.node.connection.client.create_signed_extrinsic(
                call=call, keypair=keypair
            )
        except Exception as exception:
            logging.exception(exception)
            raise NodeSmithsException(exception)

        try:
            # fixme: code stuck infinitely if no blocks are created on blockchain
            #       should have a timeout option
            result = self.node.connection.client.submit_extrinsic(
                extrinsic, wait_for_inclusion=True
            )
            logging.debug(
                "Extrinsic '%s' sent and included in block '%s'",
                result.extrinsic_hash,
                result.block_hash,
            )
        except Exception as exception:
            logging.exception(exception)
            raise NodeSmithsException(exception)

        if result.is_success is False:
            logging.error(result.error_message)
            raise NodeSmithsException(result.error_message["name"])

    def accept_invitation(self, keypair: Keypair) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NodeSmithsInterface.accept_invitation.__doc__
        )
        if (
            not self.node.connection.is_connected()
            or self.node.connection.client is None
        ):
            raise NodeSmithsException(NetworkConnectionError())

        try:
            call = self.node.connection.client.compose_call(
                call_module="SmithMembers",
                call_function="accept_invitation",
                call_params=None,
            )
        except Exception as exception:
            logging.exception(exception)
            raise NodeSmithsException(exception)

        try:
            extrinsic = self.node.connection.client.create_signed_extrinsic(
                call=call, keypair=keypair
            )
        except Exception as exception:
            logging.exception(exception)
            raise NodeSmithsException(exception)

        try:
            # fixme: code stuck infinitely if no blocks are created on blockchain
            #       should have a timeout option
            result = self.node.connection.client.submit_extrinsic(
                extrinsic, wait_for_inclusion=True
            )
            logging.debug(
                "Extrinsic '%s' sent and included in block '%s'",
                result.extrinsic_hash,
                result.block_hash,
            )
        except Exception as exception:
            logging.exception(exception)
            raise NodeSmithsException(exception)

        if result.is_success is False:
            raise NodeSmithsException(result.error_message["name"])

    def certify(self, keypair: Keypair, identity_index: int) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NodeSmithsInterface.certify.__doc__
        )
        if (
            not self.node.connection.is_connected()
            or self.node.connection.client is None
        ):
            raise NodeSmithsException(NetworkConnectionError())

        params = {
            "receiver": identity_index,
        }

        try:
            call = self.node.connection.client.compose_call(
                call_module="SmithMembers",
                call_function="certify_smith",
                call_params=params,
            )
        except Exception as exception:
            logging.exception(exception)
            raise NodeSmithsException(exception)

        try:
            extrinsic = self.node.connection.client.create_signed_extrinsic(
                call=call, keypair=keypair
            )
        except Exception as exception:
            logging.exception(exception)
            raise NodeSmithsException(exception)

        # fixme: code stuck infinitely if no blocks are created on blockchain
        #       should have a timeout option
        result = self.node.connection.client.submit_extrinsic(
            extrinsic, wait_for_inclusion=True
        )
        logging.debug(
            "Extrinsic '%s' sent and included in block '%s'",
            result.extrinsic_hash,
            result.block_hash,
        )
        if result.is_success is False:
            logging.error(result.error_message)
            raise NodeSmithsException(result.error_message["name"])

    def get_datetime_from_epoch(self, epoch_index: int) -> Optional[datetime]:
        """
        Return a datetime object from an epoch index

        :param epoch_index: Epoch number
        :return:
        """
        if (
            not self.node.connection.is_connected()
            or self.node.connection.client is None
        ):
            raise NodeSmithsException(NetworkConnectionError())

        try:
            result = self.node.connection.client.get_constant("Babe", "EpochDuration")
        except Exception as exception:
            logging.exception(exception)
            raise NodeSmithsException(exception)

        epoch_duration_in_blocks = result

        try:
            result = self.node.connection.client.get_constant(
                "Babe", "ExpectedBlockTime"
            )
        except Exception as exception:
            logging.exception(exception)
            raise NodeSmithsException(exception)

        block_duration_in_ms = result
        epoch_duration_in_ms = epoch_duration_in_blocks * block_duration_in_ms

        try:
            current_epoch_result = self.node.connection.client.query(
                "Babe", "EpochIndex"
            )
        except Exception as exception:
            logging.exception(exception)
            raise NodeSmithsException(exception)
        current_epoch_index = current_epoch_result

        current_time = datetime.now()
        epoch_diff = epoch_index - current_epoch_index
        if epoch_diff < 0:
            block_time = current_time - timedelta(
                milliseconds=abs(epoch_diff) * epoch_duration_in_ms
            )
        else:
            block_time = current_time + timedelta(
                milliseconds=abs(epoch_diff) * epoch_duration_in_ms
            )

        return block_time



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Image class for QR code
from typing import Any

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QImage, QPainter, QPixmap
from qrcode import QRCode
from qrcode.image.base import BaseImage


class QRCodeImage(BaseImage):

    # constructor
    def __init__(self, border, width, box_size, *args, **kwargs):

        # assigning border
        super().__init__(border, width, box_size, *args, **kwargs)
        self.border = border

        # assigning width
        self.width = width

        # assigning box size
        self.box_size = box_size

        # creating size
        size = (width + border * 2) * box_size

        # image
        self._image = QImage(size, size, QImage.Format_RGB16)

        # initial image as white
        self._image.fill(Qt.white)

    # pixmap method
    def pixmap(self):

        # returns image
        return QPixmap.fromImage(self._image)

    def drawrect_context(self, row: int, col: int, qr: QRCode):
        pass

    # drawrect method for drawing rectangle
    def drawrect(self, row, col):
        # creating painter object
        painter = QPainter(self._image)

        # drawing rectangle
        painter.fillRect(
            (col + self.border) * self.box_size,
            (row + self.border) * self.box_size,
            self.box_size,
            self.box_size,
            Qt.black,
        )

    def save(self, stream, kind=None):
        pass

    def new_image(self, **kwargs) -> Any:
        pass

    def process(self):
        pass



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import sys
from typing import Optional

from PyQt5 import QtCore, QtWidgets
from PyQt5.QtGui import QKeyEvent
from PyQt5.QtWidgets import QApplication, QDialog, QWidget

from tikka.domains.application import Application
from tikka.domains.entities.constants import DATA_PATH
from tikka.domains.entities.events import IndexersEvent
from tikka.domains.entities.indexer import Indexer
from tikka.slots.pyqt.resources.gui.windows.indexer_add_rc import Ui_indexerAddDialog


class IndexerAddWindow(QDialog, Ui_indexerAddDialog):
    """
    IndexerAddWindow class
    """

    def __init__(self, application: Application, parent: Optional[QWidget] = None):
        """
        Init add indexer window

        :param application: Application instance
        :param parent: QWidget instance
        """
        super().__init__(parent=parent)
        self.setupUi(self)

        self.application = application
        self._ = self.application.translator.gettext

        # buttons
        self.buttonBox.button(self.buttonBox.Ok).setEnabled(False)

        # events
        self.urlLineEdit.keyPressEvent = self.on_url_line_edit_key_press_event
        self.testButton.clicked.connect(self._on_test_button_clicked)
        self.buttonBox.accepted.connect(self.on_accepted_button)
        self.buttonBox.rejected.connect(self.close)

    def on_url_line_edit_key_press_event(self, event: QKeyEvent):
        """
        Triggered when enter is pressed to validate url in url t line edit

        :param event: QKeyEvent instance
        :return:
        """
        if event.key() == QtCore.Qt.Key_Return:
            self._on_test_button_clicked()
        else:
            QtWidgets.QLineEdit.keyPressEvent(self.urlLineEdit, event)
            # if the key is not return, handle normally

    def _on_test_button_clicked(self):
        """
        Run when use click test button

        :return:
        """
        url = self.urlLineEdit.text()
        indexer = self.application.indexers.network_test_and_get_indexer(url)
        if indexer is not None:
            self.blockValueLabel.setText(str(indexer.block))
            self.errorLabel.setText("")
            self.buttonBox.button(self.buttonBox.Ok).setEnabled(True)
        else:
            self.blockValueLabel.setText("")
            self.errorLabel.setText(self._("Impossible to connect"))
            self.buttonBox.button(self.buttonBox.Ok).setEnabled(False)

    def on_accepted_button(self) -> None:
        """
        Triggered when user click on ok button

        :return:
        """
        url = self.urlLineEdit.text()
        if self.application.indexers.get(url) is not None:
            return

        self.application.indexers.add(
            Indexer(
                url=self.urlLineEdit.text(),
                block=int(self.blockValueLabel.text()),
            )
        )

        self.application.event_dispatcher.dispatch_event(
            IndexersEvent(IndexersEvent.EVENT_TYPE_LIST_CHANGED)
        )


if __name__ == "__main__":
    qapp = QApplication(sys.argv)
    application_ = Application(DATA_PATH)
    IndexerAddWindow(application_).exec_()



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
import sys
from typing import Optional

from PyQt5.QtCore import QTimer
from PyQt5.QtGui import QFont
from PyQt5.QtWidgets import QApplication, QDialog, QLineEdit, QMessageBox, QWidget

from tikka.domains.application import Application
from tikka.domains.entities.account import Account, AccountCryptoType
from tikka.domains.entities.constants import DATA_PATH, WALLETS_PASSWORD_LENGTH
from tikka.libs.keypair import Keypair
from tikka.libs.secret import generate_alphabetic
from tikka.libs.signing_key_v1 import SigningKey
from tikka.slots.pyqt.entities.constants import (
    ADDRESS_MONOSPACE_FONT_NAME,
    DEBOUNCE_TIME,
)
from tikka.slots.pyqt.resources.gui.windows.v1_wallet_restore_rc import (
    Ui_V1WalletRestoreDialog,
)


class V1WalletRestoreWindow(QDialog, Ui_V1WalletRestoreDialog):
    """
    V1WalletRestoreWindow class
    """

    def __init__(
        self,
        application: Application,
        account: Account,
        reset_password: bool = False,
        parent: Optional[QWidget] = None,
    ):
        """
        Init V1 wallet restore window

        :param application: Application instance
        :param account: Account instance
        :param reset_password: Reset password if True (default to False)
        :param parent: QWidget instance
        """
        super().__init__(parent=parent)
        self.setupUi(self)

        self.application = application
        self.account = account
        self.reset_password = reset_password
        self._ = self.application.translator.gettext

        # set monospace font to address fields
        monospace_font = QFont(ADDRESS_MONOSPACE_FONT_NAME)
        monospace_font.setStyleHint(QFont.Monospace)
        self.addressValueLabel.setFont(monospace_font)
        self.v1AddressValueLabel.setFont(monospace_font)

        # buttons
        self.buttonBox.button(self.buttonBox.Ok).setEnabled(False)

        # events
        self.secretIDLineEdit.textChanged.connect(self._on_secret_id_line_edit_changed)
        self.passwordIDLineEdit.textChanged.connect(
            self._on_password_id_line_edit_changed
        )
        self.showButton.clicked.connect(self.on_show_button_clicked)
        self.passwordChangeButton.clicked.connect(self._generate_wallet_password)
        self.buttonBox.accepted.connect(self.on_accepted_button)
        self.buttonBox.rejected.connect(self.close)

        # debounce timer on self._generate_address()
        self.debounce_timer = QTimer()
        self.debounce_timer.timeout.connect(self._generate_address)

        # fill form
        self.storedPasswordFrame.hide()
        self.addressValueLabel.setText(self.account.address)
        self.nameValueLabel.setText(self.account.name)
        self._generate_wallet_password()

    def _on_secret_id_line_edit_changed(self):
        """
        Triggered when text is changed in the secret ID field

        :return:
        """
        if self.debounce_timer.isActive():
            self.debounce_timer.stop()
        self.debounce_timer.start(DEBOUNCE_TIME)

    def _on_password_id_line_edit_changed(self):
        """
        Triggered when text is changed in the password ID field

        :return:
        """
        if self.debounce_timer.isActive():
            self.debounce_timer.stop()
        self.debounce_timer.start(DEBOUNCE_TIME)

    def _generate_address(self):
        """
        Generate address from ID

        :return:
        """
        # stop debounce_timer to avoid infinite loop
        if self.debounce_timer.isActive():
            self.debounce_timer.stop()

        self.v1AddressValueLabel.setText("")
        self.errorLabel.setText("")
        self.buttonBox.button(self.buttonBox.Ok).setEnabled(False)

        secret_id = self.secretIDLineEdit.text().strip()
        password_id = self.passwordIDLineEdit.text().strip()
        if secret_id == "" or password_id == "":
            return

        signing_key = SigningKey.from_credentials(secret_id, password_id)
        try:
            keypair = Keypair.create_from_seed(
                seed_hex=signing_key.seed.hex(),
                ss58_format=self.application.currencies.get_current().ss58_format,
                crypto_type=AccountCryptoType.ED25519,
            )
        except Exception as exception:
            logging.exception(exception)
            self.errorLabel.setText(self._("Error generating account wallet!"))
            return

        address = keypair.ss58_address

        # if credentials address is not account address...
        if address != self.account.address:
            self.errorLabel.setText(
                self._("Generated address is not the account address!")
            )
            self.buttonBox.button(self.buttonBox.Ok).setEnabled(False)
            return

        if self.application.passwords.exists(address) and self.reset_password is False:
            stored_password = self.application.passwords.get_clear_password(keypair)
            self.storedpasswordLineEdit.setText(stored_password)
            self.storedPasswordFrame.show()
            self.passwordFrame.hide()
        else:
            self.storedPasswordFrame.hide()
            self.passwordFrame.show()

        self.v1AddressValueLabel.setText(
            Account(address).get_v1_address(
                self.application.currencies.get_current().ss58_format
            )
        )
        self.buttonBox.button(self.buttonBox.Ok).setEnabled(True)

    def on_show_button_clicked(self):
        """
        Triggered when user click on show button

        :return:
        """
        if self.secretIDLineEdit.echoMode() == QLineEdit.Password:
            self.secretIDLineEdit.setEchoMode(QLineEdit.Normal)
            self.passwordIDLineEdit.setEchoMode(QLineEdit.Normal)
            self.storedpasswordLineEdit.setEchoMode(QLineEdit.Normal)
            self.showButton.setText(self._("Hide"))
        else:
            self.secretIDLineEdit.setEchoMode(QLineEdit.Password)
            self.passwordIDLineEdit.setEchoMode(QLineEdit.Password)
            self.storedpasswordLineEdit.setEchoMode(QLineEdit.Password)
            self.showButton.setText(self._("Show"))

    def _generate_wallet_password(self):
        """
        Generate new password for wallet encryption in UI

        :return:
        """
        self.passwordLineEdit.setText(generate_alphabetic(WALLETS_PASSWORD_LENGTH))

    def on_accepted_button(self):
        """
        Triggered when user click on ok button

        :return:
        """
        secret_id = self.secretIDLineEdit.text().strip()
        password_id = self.passwordIDLineEdit.text().strip()
        password = self.passwordLineEdit.text()
        signing_key = SigningKey.from_credentials(secret_id, password_id)

        keypair = Keypair.create_from_seed(
            seed_hex=signing_key.seed.hex(),
            ss58_format=self.application.currencies.get_current().ss58_format,
            crypto_type=AccountCryptoType.ED25519,
        )
        address = keypair.ss58_address

        # if password exists for root account...
        if self.application.passwords.exists(address) and self.reset_password is False:
            # get stored password
            clear_password = self.application.passwords.get_clear_password(keypair)
            if clear_password is not None:
                password = clear_password
        else:
            # store new password
            self.application.passwords.new(keypair, password)

        wallet = self.application.wallets.get(address)
        if wallet is None:
            # create and store Wallet instance
            wallet = self.application.wallets.create(keypair, password)
            self.application.wallets.add(wallet)
        else:
            # display confirm dialog and get response
            button = QMessageBox.question(
                self,
                self._("Change wallet password?"),
                self._(
                    "A wallet already exists for this account. Change password for {address} wallet?"
                ).format(address=wallet.address),
            )
            if button == QMessageBox.Yes:
                # create and store Wallet instance
                new_wallet = self.application.wallets.create(keypair, password)
                self.application.wallets.update(new_wallet)

        self.account.file_import = False
        self.application.accounts.unlock(self.account, password)


if __name__ == "__main__":
    qapp = QApplication(sys.argv)
    application_ = Application(DATA_PATH)
    account_ = Account(
        "5GAT6CJW8yVKwUuQc7sM5Kk9GZVTpbZYk9PfjNXtvnNgAJZ1",
        name="test name",
        crypto_type=AccountCryptoType.ED25519,
        file_import=True,
    )
    V1WalletRestoreWindow(application_, account_).exec_()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'tikka/slots/pyqt/resources/gui/widgets/account_tree.ui'
#
# Created by: PyQt5 UI code generator 5.15.10
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_AccountTreeWidget(object):
    def setupUi(self, AccountTreeWidget):
        AccountTreeWidget.setObjectName("AccountTreeWidget")
        AccountTreeWidget.resize(664, 447)
        self.verticalLayout = QtWidgets.QVBoxLayout(AccountTreeWidget)
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.addCategoryButton = QtWidgets.QPushButton(AccountTreeWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.addCategoryButton.sizePolicy().hasHeightForWidth())
        self.addCategoryButton.setSizePolicy(sizePolicy)
        self.addCategoryButton.setObjectName("addCategoryButton")
        self.horizontalLayout.addWidget(self.addCategoryButton)
        self.refreshButton = QtWidgets.QPushButton(AccountTreeWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.refreshButton.sizePolicy().hasHeightForWidth())
        self.refreshButton.setSizePolicy(sizePolicy)
        self.refreshButton.setObjectName("refreshButton")
        self.horizontalLayout.addWidget(self.refreshButton, 0, QtCore.Qt.AlignRight)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.errorLabel = QtWidgets.QLabel(AccountTreeWidget)
        self.errorLabel.setStyleSheet("color: red")
        self.errorLabel.setText("")
        self.errorLabel.setAlignment(QtCore.Qt.AlignCenter)
        self.errorLabel.setObjectName("errorLabel")
        self.horizontalLayout_2.addWidget(self.errorLabel)
        self.totalBalanceValueLabel = QtWidgets.QLabel(AccountTreeWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.totalBalanceValueLabel.sizePolicy().hasHeightForWidth())
        self.totalBalanceValueLabel.setSizePolicy(sizePolicy)
        self.totalBalanceValueLabel.setStyleSheet("color: white;\n"
"border : 3px solid rgb(34, 156, 244);\n"
"border-radius: 10%;\n"
"background: rgb(34, 156, 244);\n"
"\n"
"\n"
"")
        self.totalBalanceValueLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.totalBalanceValueLabel.setObjectName("totalBalanceValueLabel")
        self.horizontalLayout_2.addWidget(self.totalBalanceValueLabel)
        self.verticalLayout.addLayout(self.horizontalLayout_2)
        self.treeView = QtWidgets.QTreeView(AccountTreeWidget)
        self.treeView.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.treeView.setSizeAdjustPolicy(QtWidgets.QAbstractScrollArea.AdjustToContents)
        self.treeView.setDragEnabled(True)
        self.treeView.setDragDropOverwriteMode(True)
        self.treeView.setDragDropMode(QtWidgets.QAbstractItemView.InternalMove)
        self.treeView.setAlternatingRowColors(True)
        self.treeView.setIndentation(20)
        self.treeView.setItemsExpandable(True)
        self.treeView.setHeaderHidden(True)
        self.treeView.setExpandsOnDoubleClick(True)
        self.treeView.setObjectName("treeView")
        self.verticalLayout.addWidget(self.treeView)

        self.retranslateUi(AccountTreeWidget)
        QtCore.QMetaObject.connectSlotsByName(AccountTreeWidget)
        AccountTreeWidget.setTabOrder(self.addCategoryButton, self.treeView)

    def retranslateUi(self, AccountTreeWidget):
        #  _translate = QtCore.QCoreApplication.translate
        AccountTreeWidget.setWindowTitle(_("Accounts"))
        self.addCategoryButton.setText(_("Add category"))
        self.refreshButton.setText(_("Refresh"))
        self.totalBalanceValueLabel.setText(_("?"))
import tikka.slots.pyqt.resources.icons.index_rc



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
from typing import Dict, List

from tikka.domains.entities.authorities import AuthorityStatus
from tikka.interfaces.adapters.network.connection import NetworkConnectionError
from tikka.interfaces.adapters.network.node.authorities import (
    NodeAuthoritiesException,
    NodeAuthoritiesInterface,
)
from tikka.libs.keypair import Keypair


class NodeAuthorities(NodeAuthoritiesInterface):
    """
    NodeAuthorities class
    """

    def rotate_keys(self) -> str:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NodeAuthoritiesInterface.rotate_keys.__doc__
        )
        if (
            not self.node.connection.is_connected()
            or self.node.connection.client is None
        ):
            raise NodeAuthoritiesException(NetworkConnectionError())

        try:
            result = self.node.connection.client.rpc_request("author_rotateKeys", [])
        except Exception as exception:
            logging.exception(exception)
            if len(exception.args) > 0 and isinstance(exception.args[0], dict):
                exception = exception.args[0]["message"]
            raise NodeAuthoritiesException(exception)

        if result is None:
            raise NodeAuthoritiesException("No result from author_rotateKeys")

        return result.get("result")  # type: ignore

    def has_session_keys(self, session_keys: str) -> bool:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NodeAuthoritiesInterface.has_session_keys.__doc__
        )
        if (
            not self.node.connection.is_connected()
            or self.node.connection.client is None
        ):
            raise NodeAuthoritiesException(NetworkConnectionError())

        try:
            result = self.node.connection.client.rpc_request(
                "author_hasSessionKeys", [session_keys]
            ).get("result")
        except Exception as exception:
            logging.exception(exception)
            raise NodeAuthoritiesException(exception)

        return result  # type: ignore

    def publish_session_keys(self, keypair: Keypair, session_keys: str) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NodeAuthoritiesInterface.publish_session_keys.__doc__
        )
        if (
            not self.node.connection.is_connected()
            or self.node.connection.client is None
        ):
            raise NodeAuthoritiesException(NetworkConnectionError())

        session_keys_bytearray = bytearray.fromhex(session_keys[2:])
        params = {
            "keys": {
                "grandpa": f"0x{session_keys_bytearray[0:32].hex()}",
                "babe": f"0x{session_keys_bytearray[32:64].hex()}",
                "im_online": f"0x{session_keys_bytearray[64:96].hex()}",
                "authority_discovery": f"0x{session_keys_bytearray[96:128].hex()}",
            }
        }

        try:
            call = self.node.connection.client.compose_call(
                call_module="AuthorityMembers",
                call_function="set_session_keys",
                call_params=params,
            )
        except Exception as exception:
            logging.exception(exception)
            raise NodeAuthoritiesException(exception)

        try:
            extrinsic = self.node.connection.client.create_signed_extrinsic(
                call=call, keypair=keypair
            )
        except Exception as exception:
            logging.exception(exception)
            raise NodeAuthoritiesException(exception)

        try:
            # fixme: code stuck infinitely if no blocks are created on blockchain
            #       should have a timeout option
            result = self.node.connection.client.submit_extrinsic(
                extrinsic, wait_for_inclusion=True
            )
        except Exception as exception:
            logging.exception(exception)
            raise NodeAuthoritiesException(exception)

        logging.debug(
            "Extrinsic '%s' sent and included in block '%s'",
            result.extrinsic_hash,
            result.block_hash,
        )

        if result.is_success is False:
            logging.error(result.error_message)
            raise NodeAuthoritiesException(result.error_message["name"])

    def go_online(self, keypair: Keypair) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NodeAuthoritiesInterface.go_online.__doc__
        )
        if (
            not self.node.connection.is_connected()
            or self.node.connection.client is None
        ):
            raise NodeAuthoritiesException(NetworkConnectionError())

        try:
            call = self.node.connection.client.compose_call(
                call_module="AuthorityMembers",
                call_function="go_online",
                call_params=None,
            )
        except Exception as exception:
            logging.exception(exception)
            raise NodeAuthoritiesException(exception)

        try:
            extrinsic = self.node.connection.client.create_signed_extrinsic(
                call=call, keypair=keypair
            )
        except Exception as exception:
            logging.exception(exception)
            raise NodeAuthoritiesException(exception)

        try:
            # fixme: code stuck infinitely if no blocks are created on blockchain
            #       should have a timeout option
            result = self.node.connection.client.submit_extrinsic(
                extrinsic, wait_for_inclusion=True
            )
            logging.debug(
                "Extrinsic '%s' sent and included in block '%s'",
                result.extrinsic_hash,
                result.block_hash,
            )
        except Exception as exception:
            logging.exception(exception)
            raise NodeAuthoritiesException(exception)

        if result.is_success is False:
            logging.error(result.error_message)
            raise NodeAuthoritiesException(result.error_message["name"])

    def go_offline(self, keypair: Keypair) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NodeAuthoritiesInterface.go_offline.__doc__
        )
        if (
            not self.node.connection.is_connected()
            or self.node.connection.client is None
        ):
            raise NodeAuthoritiesException(NetworkConnectionError())

        try:
            call = self.node.connection.client.compose_call(
                call_module="AuthorityMembers",
                call_function="go_offline",
                call_params=None,
            )
        except Exception as exception:
            logging.exception(exception)
            raise NodeAuthoritiesException(exception)

        try:
            extrinsic = self.node.connection.client.create_signed_extrinsic(
                call=call, keypair=keypair
            )
        except Exception as exception:
            logging.exception(exception)
            raise NodeAuthoritiesException(exception)

        try:
            # fixme: code stuck infinitely if no blocks are created on blockchain
            #       should have a timeout option
            result = self.node.connection.client.submit_extrinsic(
                extrinsic, wait_for_inclusion=True
            )
            logging.debug(
                "Extrinsic '%s' sent and included in block '%s'",
                result.extrinsic_hash,
                result.block_hash,
            )
        except Exception as exception:
            logging.exception(exception)
            raise NodeAuthoritiesException(exception)

        if result.is_success is False:
            logging.error(result.error_message)
            raise NodeAuthoritiesException(result.error_message["name"])

    def get_status(self, identity_index: int) -> AuthorityStatus:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NodeAuthoritiesInterface.get_status.__doc__
        )
        if (
            not self.node.connection.is_connected()
            or self.node.connection.client is None
        ):
            raise NodeAuthoritiesException(NetworkConnectionError())

        status = AuthorityStatus.OFFLINE
        storage_functions: list = [
            ("AuthorityMembers", "OnlineAuthorities", []),
            ("AuthorityMembers", "IncomingAuthorities", []),
            ("AuthorityMembers", "OutgoingAuthorities", []),
        ]
        try:
            multi_result = self.node.connection.client.query_multi(storage_functions)
        except Exception as exception:
            logging.exception(exception)
            raise NodeAuthoritiesException(exception)

        online_authorities = multi_result[0] or []
        incoming_authorities = multi_result[1] or []
        outgoing_authorities = multi_result[2] or []

        if identity_index in online_authorities:
            status = AuthorityStatus.ONLINE
        elif identity_index in incoming_authorities:
            status = AuthorityStatus.INCOMING
        elif identity_index in outgoing_authorities:
            status = AuthorityStatus.OUTGOING

        return status

    def get_all(self) -> Dict[int, List[int]]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NodeAuthoritiesInterface.get_all.__doc__
        )
        all_by_status = {}

        if (
            not self.node.connection.is_connected()
            or self.node.connection.client is None
        ):
            raise NodeAuthoritiesException(NetworkConnectionError())

        storage_functions: list = [
            ("AuthorityMembers", "OnlineAuthorities", []),
            ("AuthorityMembers", "IncomingAuthorities", []),
            ("AuthorityMembers", "OutgoingAuthorities", []),
        ]
        try:
            multi_result = self.node.connection.client.query_multi(storage_functions)
        except Exception as exception:
            logging.exception(exception)
            raise NodeAuthoritiesException(exception)

        all_by_status[AuthorityStatus.ONLINE.value] = multi_result[0] or []
        all_by_status[AuthorityStatus.INCOMING.value] = multi_result[1] or []
        all_by_status[AuthorityStatus.OUTGOING.value] = multi_result[2] or []

        return all_by_status



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
from typing import Optional

from PyQt5.QtCore import QMutex
from PyQt5.QtWidgets import QApplication, QDialog, QWidget

from tikka.domains.application import Application
from tikka.domains.entities.constants import DATA_PATH
from tikka.slots.pyqt.resources.gui.windows.welcome_rc import Ui_WelcomeDialog
from tikka.slots.pyqt.windows.account_create import AccountCreateWindow
from tikka.slots.pyqt.windows.account_import import AccountImportWindow
from tikka.slots.pyqt.windows.v1_account_import import V1AccountImportWindow


class WelcomeWindow(QDialog, Ui_WelcomeDialog):
    """
    WelcomenWindow class
    """

    def __init__(self, application: Application, parent: Optional[QWidget] = None):
        """
        Init welcome window

        :param application: Application instance
        :param parent: QWidget instance
        """
        super().__init__(parent=parent)
        self.setupUi(self)

        self.application = application
        self._ = self.application.translator.gettext

        # events
        self.importV2Button.clicked.connect(self.on_import_v2_button_clicked)
        self.importV1Button.clicked.connect(self.on_import_v1_button_clicked)
        self.createButton.clicked.connect(self.on_create_button_clicked)
        self.continueButton.clicked.connect(self.on_continue_button_clicked)

    def on_import_v2_button_clicked(self):
        """
        Triggered when user click on import V2 account button

        :return:
        """
        window = AccountImportWindow(self.application, self.parentWidget())
        window.destroyed.connect(self.close)
        window.exec()

    def on_import_v1_button_clicked(self):
        """
        Triggered when user click on import V1 account button

        :return:
        """
        window = V1AccountImportWindow(self.application, QMutex(), self.parentWidget())
        window.destroyed.connect(self.close)
        window.exec()

    def on_create_button_clicked(self):
        """
        Triggered when user click on create account button

        :return:
        """
        window = AccountCreateWindow(self.application, self.parentWidget())
        window.destroyed.connect(self.close)
        window.exec()

    def on_continue_button_clicked(self):
        """
        Triggered when user click on continue button

        :return:
        """
        self.close()


if __name__ == "__main__":
    qapp = QApplication(sys.argv)
    application_ = Application(DATA_PATH)
    WelcomeWindow(application_).exec_()



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import List, Optional

from tikka.domains.entities.authorities import Authority, AuthorityStatus
from tikka.domains.entities.node import Node
from tikka.domains.nodes import Nodes
from tikka.domains.smiths import Smiths
from tikka.interfaces.adapters.network.node.authorities import NodeAuthoritiesInterface
from tikka.interfaces.adapters.repository.authorities import (
    AuthoritiesRepositoryInterface,
)
from tikka.libs.keypair import Keypair


class Authorities:

    """
    Authorities domain class
    """

    def __init__(
        self,
        repository: AuthoritiesRepositoryInterface,
        network: NodeAuthoritiesInterface,
        nodes: Nodes,
        smiths: Smiths,
    ):
        """
        Init Authorities domain

        :param repository: AuthoritiesRepositoryInterface instance
        :param network: NodeAuthoritiesInterface instance
        :param nodes: Nodes domain instance
        :param smiths: Smiths domain instance
        """
        self.network = network
        self.repository = repository
        self.nodes = nodes
        self.smiths = smiths

    @staticmethod
    def create(identity_index: int, status: AuthorityStatus):
        """
        Return Authority instance from parameters

        :param identity_index: Identity index
        :param status: AuthorityStatus enum
        :return:
        """
        return Authority(identity_index, status)

    def add(self, authority: Authority) -> None:
        """
        Add Authority in repository

        :param authority: Authority instance
        :return:
        """
        self.repository.add(authority)

    def get(self, identity_index: int) -> Optional[Authority]:
        """
        Get Authority by identity index from repository

        :param identity_index: Identity index
        :return:
        """
        return self.repository.get(identity_index)

    def update(self, authority: Authority):
        """
        Update authority in repository

        :param authority: Authority instance
        :return:
        """
        self.repository.update(authority)

    def delete(self, identity_index: int) -> None:
        """
        Delete Authority in repository

        :param identity_index: Identity index to delete
        :return:
        """
        self.repository.delete(identity_index)

    def delete_all(self) -> None:
        """
        Delete all authorities in repository

        :return:
        """
        self.repository.delete_all()

    def exists(self, identity_index: int) -> bool:
        """
        Return True if Authority exists in repository

        :param identity_index: Identity index to check
        :return:
        """
        return self.repository.exists(identity_index)

    def list(self) -> List[Authority]:
        """
        Return list of all Authority in repository

        :return:
        """
        return self.repository.list()

    def get_status(self, identity_index: int) -> AuthorityStatus:
        """
        Return AuthorityStatus enum for identity_index

        :param identity_index: Identity index
        :return:
        """
        if not self.repository.exists(identity_index):
            return AuthorityStatus.OFFLINE

        authority = self.repository.get(identity_index)
        if authority is None:
            return AuthorityStatus.OFFLINE

        return authority.status

    def network_rotate_keys(self, node: Node) -> Optional[str]:
        """
        Change node session keys and return them

        :param node: Node instance
        :return:
        """
        session_keys = self.network.rotate_keys()
        if session_keys is not None:
            node.session_keys = session_keys
            self.nodes.update(node)
        return session_keys

    def network_has_session_keys(self, session_keys: str) -> Optional[bool]:
        """
        Return True if the current node keystore store private session keys corresponding to public session_keys

        :param session_keys: Session public keys (hex string "0x123XYZ")
        :return:
        """
        return self.network.has_session_keys(session_keys)

    def network_publish_session_keys(self, keypair: Keypair, session_keys: str) -> None:
        """
        Set/Change in blockchain the session public keys for the Keypair account

        :param keypair: Owner Keypair
        :param session_keys: Session public keys (hex string "0x123XYZ")
        :return:
        """
        return self.network.publish_session_keys(keypair, session_keys)

    def network_go_online(self, keypair: Keypair) -> None:
        """
        Start writing blocks with smith account from keypair

        :param keypair: Smith account Keypair
        :return:
        """
        return self.network.go_online(keypair)

    def network_go_offline(self, keypair: Keypair) -> None:
        """
        Stop writing blocks with smith account from keypair

        :param keypair: Smith account Keypair
        :return:
        """
        return self.network.go_offline(keypair)

    def network_get_status(self, identity_index: int) -> AuthorityStatus:
        """
        Return AuthorityStatus enum for identity_index

        :param identity_index: Identity index
        :return:
        """
        return self.network.get_status(identity_index)

    def network_get_all(self) -> None:
        """
        Get all authorities by status

        :return:
        """
        network_authorities = self.network.get_all()

        # list of known smith
        smith_identity_indice = [smith.identity_index for smith in self.smiths.list()]
        authorities = self.list()

        # purge obsolete authorities
        for authority in authorities:
            if authority.identity_index not in smith_identity_indice:
                self.delete(authority.identity_index)

        for authority in authorities:
            if (
                authority.identity_index
                in network_authorities[AuthorityStatus.ONLINE.value]
            ):
                authority.status = AuthorityStatus.ONLINE
            elif (
                authority.identity_index
                in network_authorities[AuthorityStatus.INCOMING.value]
            ):
                authority.status = AuthorityStatus.INCOMING
            elif (
                authority.identity_index
                in network_authorities[AuthorityStatus.OUTGOING.value]
            ):
                authority.status = AuthorityStatus.OUTGOING
            else:
                authority.status = AuthorityStatus.OFFLINE
            self.update(authority)

        # new smiths
        authority_identity_indice = [
            authority.identity_index for authority in authorities
        ]
        for identity_index in smith_identity_indice:
            if identity_index not in authority_identity_indice:
                authority = self.create(identity_index, AuthorityStatus.OFFLINE)
                if (
                    authority.identity_index
                    in network_authorities[AuthorityStatus.ONLINE.value]
                ):
                    authority.status = AuthorityStatus.ONLINE
                elif (
                    authority.identity_index
                    in network_authorities[AuthorityStatus.INCOMING.value]
                ):
                    authority.status = AuthorityStatus.INCOMING
                elif (
                    authority.identity_index
                    in network_authorities[AuthorityStatus.OUTGOING.value]
                ):
                    authority.status = AuthorityStatus.OUTGOING
                self.add(authority)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'tikka/slots/pyqt/resources/gui/widgets/servers.ui'
#
# Created by: PyQt5 UI code generator 5.15.10
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_ServersWidget(object):
    def setupUi(self, ServersWidget):
        ServersWidget.setObjectName("ServersWidget")
        ServersWidget.resize(724, 526)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(ServersWidget)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.nodesGroupBox = QtWidgets.QGroupBox(ServersWidget)
        self.nodesGroupBox.setObjectName("nodesGroupBox")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.nodesGroupBox)
        self.verticalLayout.setObjectName("verticalLayout")
        self.addNodeButton = QtWidgets.QPushButton(self.nodesGroupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.addNodeButton.sizePolicy().hasHeightForWidth())
        self.addNodeButton.setSizePolicy(sizePolicy)
        self.addNodeButton.setObjectName("addNodeButton")
        self.verticalLayout.addWidget(self.addNodeButton, 0, QtCore.Qt.AlignRight)
        self.nodesTableView = QtWidgets.QTableView(self.nodesGroupBox)
        self.nodesTableView.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.nodesTableView.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.nodesTableView.setTabKeyNavigation(False)
        self.nodesTableView.setProperty("showDropIndicator", False)
        self.nodesTableView.setAlternatingRowColors(True)
        self.nodesTableView.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        self.nodesTableView.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.nodesTableView.setShowGrid(False)
        self.nodesTableView.setSortingEnabled(True)
        self.nodesTableView.setObjectName("nodesTableView")
        self.nodesTableView.horizontalHeader().setStretchLastSection(True)
        self.nodesTableView.verticalHeader().setVisible(False)
        self.verticalLayout.addWidget(self.nodesTableView)
        self.verticalLayout_2.addWidget(self.nodesGroupBox)
        self.indexersGroupBox = QtWidgets.QGroupBox(ServersWidget)
        self.indexersGroupBox.setObjectName("indexersGroupBox")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.indexersGroupBox)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.addIndexerButton = QtWidgets.QPushButton(self.indexersGroupBox)
        self.addIndexerButton.setObjectName("addIndexerButton")
        self.verticalLayout_3.addWidget(self.addIndexerButton, 0, QtCore.Qt.AlignRight)
        self.indexersTableView = QtWidgets.QTableView(self.indexersGroupBox)
        self.indexersTableView.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.indexersTableView.setTabKeyNavigation(False)
        self.indexersTableView.setProperty("showDropIndicator", False)
        self.indexersTableView.setAlternatingRowColors(True)
        self.indexersTableView.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        self.indexersTableView.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.indexersTableView.setShowGrid(False)
        self.indexersTableView.setSortingEnabled(True)
        self.indexersTableView.setObjectName("indexersTableView")
        self.indexersTableView.horizontalHeader().setStretchLastSection(True)
        self.indexersTableView.verticalHeader().setVisible(False)
        self.verticalLayout_3.addWidget(self.indexersTableView)
        self.verticalLayout_2.addWidget(self.indexersGroupBox)

        self.retranslateUi(ServersWidget)
        QtCore.QMetaObject.connectSlotsByName(ServersWidget)

    def retranslateUi(self, ServersWidget):
        #  _translate = QtCore.QCoreApplication.translate
        ServersWidget.setWindowTitle(_("Network servers"))
        self.nodesGroupBox.setTitle(_("Nodes"))
        self.addNodeButton.setText(_("Add node"))
        self.indexersGroupBox.setTitle(_("Indexers"))
        self.addIndexerButton.setText(_("Add indexer"))
import tikka.slots.pyqt.resources.icons.index_rc



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from dataclasses import dataclass
from typing import Optional


@dataclass
class Currency:

    code_name: str
    name: str
    ss58_format: int
    token_decimals: Optional[int] = None
    token_symbol: Optional[str] = None
    universal_dividend: Optional[int] = None
    monetary_mass: Optional[int] = None
    members_count: Optional[int] = None
    block_duration: int = 6000
    epoch_duration: int = 3600000

    certification_number_to_be_member: Optional[int] = None
    minimum_delay_between_two_membership_renewals: Optional[int] = None
    validity_duration_of_membership: Optional[int] = None
    minimum_certifications_received_to_be_certifier: Optional[int] = None
    validity_duration_of_certification: Optional[int] = None
    minimum_delay_between_two_certifications: Optional[int] = None
    maximum_number_of_certifications_per_member: Optional[int] = None
    maximum_distance_in_step: Optional[int] = None
    minimum_percentage_of_remote_referral_members_to_be_member: Optional[int] = None
    identity_automatic_revocation_period: Optional[int] = None
    minimum_delay_between_changing_identity_owner: Optional[int] = None
    confirm_identity_period: Optional[int] = None
    identity_deletion_after_revocation: Optional[int] = None
    minimum_delay_between_identity_creation: Optional[int] = None
    identity_validation_period: Optional[int] = None
    maximum_certifications_per_smith: Optional[int] = None
    number_of_certifications_to_become_smith: Optional[int] = None
    maximum_inactivity_duration_allowed_for_smith: Optional[int] = None



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import json
import logging
import sys
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from PyQt5 import QtGui
from PyQt5.QtCore import QModelIndex, QMutex, QSize, Qt, pyqtSignal
from PyQt5.QtGui import QColor, QIcon, QKeyEvent, QPixmap, QShowEvent
from PyQt5.QtWidgets import (
    QApplication,
    QComboBox,
    QFileDialog,
    QMainWindow,
    QPushButton,
    QWidget,
)

from tikka import __version__
from tikka.domains.application import Application
from tikka.domains.config import Config
from tikka.domains.entities.account import Account
from tikka.domains.entities.address import DisplayAddress
from tikka.domains.entities.constants import DATA_PATH
from tikka.domains.entities.events import (
    AccountEvent,
    ConnectionsEvent,
    CurrencyEvent,
    UnitEvent,
)
from tikka.slots.pyqt.entities.constants import (
    ICON_ACCOUNT_NO_WALLET,
    ICON_ACCOUNT_WALLET_LOCKED,
    ICON_ACCOUNT_WALLET_UNLOCKED,
    ICON_NETWORK_CONNECTED,
    ICON_NETWORK_DISCONNECTED,
    SAVE_DATA_DEFAULT_DIR_PREFERENCES_KEY,
    SELECTED_TAB_PAGE_PREFERENCES_KEY,
    SELECTED_UNIT_PREFERENCES_KEY,
    TABS_PREFERENCES_KEY,
)
from tikka.slots.pyqt.entities.worker import AsyncQWorker
from tikka.slots.pyqt.resources.gui.windows.main_window_rc import Ui_MainWindow
from tikka.slots.pyqt.widgets.account import AccountWidget
from tikka.slots.pyqt.widgets.account_table import AccountTableWidget
from tikka.slots.pyqt.widgets.account_tree import AccountTreeWidget
from tikka.slots.pyqt.widgets.connection import ConnectionWidget
from tikka.slots.pyqt.widgets.currency import CurrencyWidget
from tikka.slots.pyqt.widgets.licence import LicenceWidget
from tikka.slots.pyqt.widgets.servers import ServersWidget
from tikka.slots.pyqt.widgets.smith import SmithWidget
from tikka.slots.pyqt.widgets.tab_widget import TabWidget
from tikka.slots.pyqt.widgets.technical_committee import TechnicalCommitteeWidget
from tikka.slots.pyqt.windows.about import AboutWindow
from tikka.slots.pyqt.windows.account_create import AccountCreateWindow
from tikka.slots.pyqt.windows.account_import import AccountImportWindow
from tikka.slots.pyqt.windows.address_add import AddressAddWindow
from tikka.slots.pyqt.windows.configuration import ConfigurationWindow
from tikka.slots.pyqt.windows.scan_qr_code_open_cv import ScanQRCodeOpenCVWindow
from tikka.slots.pyqt.windows.transfer import TransferWindow
from tikka.slots.pyqt.windows.v1_account_import import V1AccountImportWindow
from tikka.slots.pyqt.windows.v1_account_import_wizard import (
    V1AccountImportWizardWindow,
)
from tikka.slots.pyqt.windows.v1_file_import import V1FileImportWindow
from tikka.slots.pyqt.windows.vault_import_by_mnemonic import (
    VaultImportByMnemonicWindow,
)
from tikka.slots.pyqt.windows.welcome import WelcomeWindow

if TYPE_CHECKING:
    pass


class MainWindow(QMainWindow, Ui_MainWindow):
    """
    MainWindow class
    """

    window_shown = pyqtSignal()
    repository_data_updated = pyqtSignal()
    mutex = QMutex()

    def __init__(self, application: Application, parent: Optional[QWidget] = None):
        """
        Init main window

        :param application: Application instance
        :param parent: QWidget instance
        """
        super().__init__(parent=parent)
        self.setupUi(self)

        self.application = application
        self._ = self.application.translator.gettext

        self.update_title()

        # tab widget
        self.tabWidget.close()
        self.tabWidget = TabWidget(self.application)
        self.tabWidget.setParent(self.centralwidget)
        self.tabWidget.setTabsClosable(True)
        self.tabWidget.setMovable(True)
        self.tabWidget.setObjectName("tabWidget")
        self.verticalLayout.addWidget(self.tabWidget)

        # slots
        self.tabWidget.tabCloseRequested.connect(self.close_tab)
        self.window_shown.connect(self.on_window_shown)

        # connect functions to menu actions
        # accounts menu
        self.actionTransfer.triggered.connect(self.open_transfer_window)
        self.actionAccount_tree.triggered.connect(self.add_account_tree_tab)
        self.actionAccount_table.triggered.connect(self.add_account_table_tab)
        self.actionScan_a_QRCode.triggered.connect(self.open_scan_qrcode_window)
        self.actionAdd_an_address.triggered.connect(self.open_add_address_window)
        self.actionImport_account.triggered.connect(self.open_import_account_window)
        self.actionCreate_account.triggered.connect(self.open_create_account_window)
        self.actionQuit.triggered.connect(self.close)

        # Vaults menu
        self.actionImport_by_Mnemonic.triggered.connect(
            self.open_vault_import_by_mnemonic_window
        )

        # V1 accounts menu
        self.actionImport_in_V2_wizard.triggered.connect(
            self.open_v1_import_in_v2_wizard_window
        )
        self.actionV1Import_account.triggered.connect(
            self.open_v1_import_account_window
        )
        self.actionV1Import_file.triggered.connect(self.open_v1_import_file_window)

        # network menu
        self.actionConnection.triggered.connect(self.add_connection_tab)
        self.actionServers.triggered.connect(self.add_servers_tab)

        # advanced menu
        self.actionSmith.triggered.connect(self.add_smith_tab)
        self.actionTechnical_Committee.triggered.connect(
            self.add_technical_committee_tab
        )
        self.actionSave_user_data.triggered.connect(self.open_save_user_data_window)
        self.actionRestore_user_data.triggered.connect(
            self.open_restore_user_data_window
        )
        # help menu
        self.actionWelcome.triggered.connect(self.open_welcome_window)
        self.actionCurrency.triggered.connect(self.add_currency_tab)
        self.actionG1_licence.triggered.connect(self.add_licence_tab)
        self.actionConfiguration.triggered.connect(self.open_configuration_window)
        self.actionAbout.triggered.connect(self.open_about_window)

        # status bar
        # Unit selector
        self.unit_combo_box = QComboBox()
        self.statusbar.addPermanentWidget(self.unit_combo_box)
        self.init_units()

        # Node connection status icon
        self.node_connection_status_icon = QPushButton()
        self.node_connection_status_icon.setFlat(True)
        self.node_connection_status_icon.setFixedSize(QSize(16, 16))
        self.node_connection_status_icon.setToolTip(
            self._("Node connection status. Click to open connection tab.")
        )
        self.statusbar.addPermanentWidget(self.node_connection_status_icon)

        # Indexer connection status icon
        self.indexer_connection_status_icon = QPushButton()
        self.indexer_connection_status_icon.setFlat(True)
        self.indexer_connection_status_icon.setFixedSize(QSize(16, 16))
        self.indexer_connection_status_icon.setToolTip(
            self._("Indexer connection status. Click to open connection tab.")
        )
        self.statusbar.addPermanentWidget(self.indexer_connection_status_icon)
        self.init_connection_status()

        # slots
        self.unit_combo_box.activated.connect(self._on_unit_changed)
        self.node_connection_status_icon.clicked.connect(self.add_connection_tab)
        self.indexer_connection_status_icon.clicked.connect(self.add_connection_tab)

        # application events
        self.application.event_dispatcher.add_event_listener(
            CurrencyEvent.EVENT_TYPE_PRE_CHANGE, self.on_currency_event
        )
        self.application.event_dispatcher.add_event_listener(
            CurrencyEvent.EVENT_TYPE_CHANGED, self.on_currency_event
        )
        self.application.event_dispatcher.add_event_listener(
            AccountEvent.EVENT_TYPE_ADD, self.on_add_account_event
        )
        self.application.event_dispatcher.add_event_listener(
            AccountEvent.EVENT_TYPE_DELETE, self.on_delete_account_event
        )
        self.application.event_dispatcher.add_event_listener(
            AccountEvent.EVENT_TYPE_UPDATE, self.on_update_account_event
        )
        self.application.event_dispatcher.add_event_listener(
            ConnectionsEvent.EVENT_TYPE_NODE_CONNECTED, self._on_node_connected
        )
        self.application.event_dispatcher.add_event_listener(
            ConnectionsEvent.EVENT_TYPE_NODE_DISCONNECTED, self._on_node_disconnected
        )
        self.application.event_dispatcher.add_event_listener(
            ConnectionsEvent.EVENT_TYPE_INDEXER_CONNECTED, self._on_indexer_connected
        )
        self.application.event_dispatcher.add_event_listener(
            ConnectionsEvent.EVENT_TYPE_INDEXER_DISCONNECTED,
            self._on_indexer_disconnected,
        )

        #####################################
        # ASYNC WORKERS
        #####################################
        self.connect_to_node_async_worker = AsyncQWorker(
            self.connect_to_node, self.mutex
        )
        self.connect_to_node_async_worker.finished.connect(
            self._on_finished_connect_to_node
        )
        # async connect to indexer
        self.connect_to_indexer_async_worker = AsyncQWorker(
            self.connect_to_indexer, self.mutex
        )
        self.connect_to_indexer_async_worker.finished.connect(
            self._on_finished_connect_to_indexer
        )
        # fetch data from network
        self.fetch_data_from_network_node_async_qworker = AsyncQWorker(
            self.fetch_data_from_network_node, self.mutex
        )
        self.fetch_data_from_network_node_async_qworker.finished.connect(
            self._on_finished_fetch_data_from_network_node
        )
        self.fetch_data_from_network_indexer_async_qworker = AsyncQWorker(
            self.fetch_data_from_network_indexer, self.mutex
        )
        self.fetch_data_from_network_indexer_async_qworker.finished.connect(
            self._on_finished_fetch_data_from_network_indexer
        )

        # open saved tabs
        self.init_tabs()

    def showEvent(self, event: QShowEvent):
        """
        Triggered when the window is shown

        :param event: QEvent instance
        :return:
        """
        super().showEvent(event)

        self.window_shown.emit()

    def on_window_shown(self):
        """
        Triggered when the window is shown

        :return:
        """
        self.statusbar.showMessage(self._("Connecting to servers..."), 10000)
        self.connect_to_node_async_worker.start()
        self.connect_to_indexer_async_worker.start()

    def keyPressEvent(self, event: QKeyEvent) -> None:
        """
        Triggered when user press a key

        :param event: QKeyEvent instance
        :return:
        """
        # close current tab with "ctrl-w"
        if (
            event.key() == Qt.Key_W
            and (event.modifiers() & Qt.ControlModifier) == Qt.ControlModifier
        ):
            self.close_current_tab()

    def closeEvent(
        self, event: QtGui.QCloseEvent  # pylint: disable=unused-argument
    ) -> None:
        """
        Override close event

        :param event:
        :return:
        """
        # save tabs in repository
        self.save_tabs()

        # save tab selection in preferences
        self.application.repository.preferences.set(
            SELECTED_TAB_PAGE_PREFERENCES_KEY, self.tabWidget.currentIndex()
        )

        self.application.close()

    def init_units(self) -> None:
        """
        Init units combobox in status bar

        :return:
        """
        self.unit_combo_box.clear()

        for key, amount in self.application.amounts.register.items():
            self.unit_combo_box.addItem(amount.name(), userData=key)
        preferences_selected_unit = self.application.repository.preferences.get(
            SELECTED_UNIT_PREFERENCES_KEY
        )
        if preferences_selected_unit is None:
            # set first unit in preferences
            self.application.repository.preferences.set(
                SELECTED_UNIT_PREFERENCES_KEY,
                self.application.amounts.get_register_keys()[0],
            )
            preferences_selected_unit = self.application.repository.preferences.get(
                SELECTED_UNIT_PREFERENCES_KEY
            )

        self.unit_combo_box.setCurrentIndex(
            self.unit_combo_box.findData(preferences_selected_unit)
        )

    def init_connection_status(self):
        """
        Init connection status icon

        :return:
        """
        if self.application.connections.node.is_connected():
            self.node_connection_status_icon.setIcon(
                QIcon(QPixmap(ICON_NETWORK_CONNECTED))
            )
        else:
            self.node_connection_status_icon.setIcon(
                QIcon(QPixmap(ICON_NETWORK_DISCONNECTED))
            )
        if self.application.connections.indexer.is_connected():
            self.indexer_connection_status_icon.setIcon(
                QIcon(QPixmap(ICON_NETWORK_CONNECTED))
            )
        else:
            self.indexer_connection_status_icon.setIcon(
                QIcon(QPixmap(ICON_NETWORK_DISCONNECTED))
            )

    def init_tabs(self):
        """
        Init tabs from repository

        :return:
        """
        # close all tabs
        self.tabWidget.clear()

        result = self.application.preferences.get(TABS_PREFERENCES_KEY)
        if result is None:
            return

        opened_tabs: Dict[str, str] = json.loads(result)

        # fetch tabs from repository
        for title, panel_class in opened_tabs.items():
            # if account tab...
            if panel_class == AccountWidget.__name__:
                # get account
                account = self.application.accounts.get_by_address(title)
                if account is not None:
                    self.add_account_tab(account)
            elif panel_class == CurrencyWidget.__name__:
                self.add_currency_tab()
            elif panel_class == LicenceWidget.__name__:
                self.add_licence_tab()
            elif panel_class == AccountTreeWidget.__name__:
                self.add_account_tree_tab()
            elif panel_class == AccountTableWidget.__name__:
                self.add_account_table_tab()
            elif panel_class == ConnectionWidget.__name__:
                self.add_connection_tab()
            elif panel_class == ServersWidget.__name__:
                self.add_servers_tab()
            elif panel_class == SmithWidget.__name__:
                self.add_smith_tab()
            elif panel_class == TechnicalCommitteeWidget.__name__:
                self.add_technical_committee_tab()

        # get preferences
        preferences_selected_page = self.application.repository.preferences.get(
            SELECTED_TAB_PAGE_PREFERENCES_KEY
        )
        if preferences_selected_page is not None:
            self.tabWidget.setCurrentIndex(int(preferences_selected_page))

    def save_tabs(self):
        """
        Save opened tabs in preferences

        :return:
        """
        # save tabwidget tabs in repository
        tabs: Dict[str, str] = {}
        for index in range(0, self.tabWidget.count()):
            widget = self.tabWidget.widget(index)
            if isinstance(widget, AccountWidget):
                # save account tab in repository
                tabs[widget.account.address] = str(widget.__class__.__name__)
            else:
                tabs[str(widget.__class__.__name__)] = str(widget.__class__.__name__)

        self.application.preferences.set(TABS_PREFERENCES_KEY, json.dumps(tabs))

    def close_tab(self, index: int):
        """
        Close tab on signal

        :param index: Index of tab requested to close
        :return:
        """
        self.tabWidget.removeTab(index)

    def close_current_tab(self) -> None:
        """
        Close current tab

        :return:
        """
        self.tabWidget.removeTab(self.tabWidget.currentIndex())

    def add_account_tree_tab(self) -> None:
        """
        Open account tree tab

        :return:
        """
        # select account tree tab if exists
        for widget in self.get_tab_widgets_by_class(AccountTreeWidget):
            self.tabWidget.setCurrentIndex(self.get_tab_index_from_widget(widget))
            return

        # create tab
        account_tree_widget = AccountTreeWidget(
            self.application, self.mutex, self.tabWidget
        )
        self.tabWidget.addTab(account_tree_widget, self._("Account tree"))
        # catch account tree double click signal
        account_tree_widget.treeView.doubleClicked.connect(
            self.on_account_tree_double_click
        )
        self.repository_data_updated.connect(
            account_tree_widget._on_finished_fetch_from_network
        )

        self.tabWidget.setCurrentIndex(self.tabWidget.count() - 1)

    def add_account_table_tab(self) -> None:
        """
        Open account table tab

        :return:
        """
        # select account table tab if exists
        for widget in self.get_tab_widgets_by_class(AccountTableWidget):
            self.tabWidget.setCurrentIndex(self.get_tab_index_from_widget(widget))
            return

        # create tab
        account_table_widget = AccountTableWidget(
            self.application, self.mutex, self.tabWidget
        )
        self.tabWidget.addTab(account_table_widget, self._("Account table"))
        # catch account list double click signal
        account_table_widget.tableView.doubleClicked.connect(
            self.on_account_table_double_click
        )
        self.repository_data_updated.connect(
            account_table_widget._on_finished_fetch_from_network
        )

        self.tabWidget.setCurrentIndex(self.tabWidget.count() - 1)

    def add_account_tab(self, account: Account) -> AccountWidget:
        """
        Open account list tab

        :return:
        """
        # select account tab if exists
        for widget in self.get_tab_widgets_by_class(AccountWidget):
            if (
                isinstance(widget, AccountWidget)
                and widget.account.address == account.address
            ):
                self.tabWidget.setCurrentIndex(self.get_tab_index_from_widget(widget))
                return widget
        account_widget = AccountWidget(
            self.application, account, self.mutex, self.tabWidget
        )
        index = self.tabWidget.addTab(
            account_widget,
            self.get_account_tab_icon(account),
            account.name,
        )
        self.set_account_tab_title(index, account)

        self.repository_data_updated.connect(
            account_widget._on_finished_fetch_from_network
        )

        self.tabWidget.setCurrentIndex(index)

        return account_widget

    def set_account_tab_title(self, index: int, account: Account):
        """
        Set account tab title with text, font and color

        :param index: Index of tab
        :param account: Account instance
        :return:
        """
        set_identity_name_color = False
        if account.name is None:
            identity = self.application.identities.get_by_address(account.address)
            if identity is None:
                tab_title = DisplayAddress(account.address).shorten
            else:
                identity_name = identity.name or ""
                tab_title = f"{identity_name}#{identity.index}"
                set_identity_name_color = True
        else:
            tab_title = account.name
            set_identity_name_color = False

        self.tabWidget.setTabText(index, tab_title)

        if set_identity_name_color is True:
            self.tabWidget.tabBar().setTabTextColor(index, QColor("blue"))
        else:
            self.tabWidget.tabBar().setTabTextColor(index, QColor("black"))

    def get_account_tab_icon(self, account: Account) -> QIcon:
        """
        Return QIcon instance for account tab icon

        :param account: Account instance
        :return:
        """
        if self.application.wallets.exists(account.address):
            if self.application.wallets.is_unlocked(account.address) is True:
                icon = QIcon(ICON_ACCOUNT_WALLET_UNLOCKED)
            else:
                icon = QIcon(ICON_ACCOUNT_WALLET_LOCKED)
        else:
            icon = QIcon(ICON_ACCOUNT_NO_WALLET)

        return icon

    def add_currency_tab(self):
        """
        Open currency tab

        :return:
        """
        # select currency tab if exists
        for widget in self.get_tab_widgets_by_class(CurrencyWidget):
            self.tabWidget.setCurrentIndex(self.get_tab_index_from_widget(widget))
            return

        currency_widget = CurrencyWidget(self.application, self.mutex, self.tabWidget)

        # create tab
        self.tabWidget.addTab(
            currency_widget,
            self._("Currency"),
        )
        self.repository_data_updated.connect(
            currency_widget._on_finished_fetch_from_network
        )

        self.tabWidget.setCurrentIndex(self.tabWidget.count() - 1)

    def add_licence_tab(self):
        """
        Open licence tab

        :return:
        """
        # select tab if exists
        for widget in self.get_tab_widgets_by_class(LicenceWidget):
            self.tabWidget.setCurrentIndex(self.get_tab_index_from_widget(widget))
            return

        # create tab
        self.tabWidget.addTab(
            LicenceWidget(self.application, self.tabWidget), self._("Ğ1 licence")
        )
        self.tabWidget.setCurrentIndex(self.tabWidget.count() - 1)

    def add_connection_tab(self):
        """
        Open network connection tab

        :return:
        """
        # select tab if exists
        for widget in self.get_tab_widgets_by_class(ConnectionWidget):
            self.tabWidget.setCurrentIndex(self.get_tab_index_from_widget(widget))
            return

        connection_widget = ConnectionWidget(
            self.application, self.mutex, self.tabWidget
        )
        # create tab
        self.tabWidget.addTab(
            connection_widget,
            self._("Connection"),
        )
        self.repository_data_updated.connect(
            connection_widget._on_finished_fetch_node_from_network
        )
        self.repository_data_updated.connect(
            connection_widget._on_finished_fetch_indexer_from_network
        )

        self.tabWidget.setCurrentIndex(self.tabWidget.count() - 1)

    def add_servers_tab(self):
        """
        Open network servers tab

        :return:
        """
        # select tab if exists
        for widget in self.get_tab_widgets_by_class(ServersWidget):
            self.tabWidget.setCurrentIndex(self.get_tab_index_from_widget(widget))
            return

        # create tab
        self.tabWidget.addTab(
            ServersWidget(self.application, self.mutex, self.tabWidget),
            self._("Servers"),
        )
        self.tabWidget.setCurrentIndex(self.tabWidget.count() - 1)

    def add_smith_tab(self):
        """
        Open smith tab

        :return:
        """
        # select smith tab if exists
        for widget in self.get_tab_widgets_by_class(SmithWidget):
            self.tabWidget.setCurrentIndex(self.get_tab_index_from_widget(widget))
            return

        smith_widget = SmithWidget(self.application, self.mutex, self.tabWidget)

        # create tab
        self.tabWidget.addTab(
            smith_widget,
            self._("Smith"),
        )

        self.repository_data_updated.connect(
            smith_widget._on_finished_fetch_all_from_network
        )

        self.tabWidget.setCurrentIndex(self.tabWidget.count() - 1)

    def add_technical_committee_tab(self):
        """
        Open technical_committee tab

        :return:
        """
        # select smith tab if exists
        for widget in self.get_tab_widgets_by_class(TechnicalCommitteeWidget):
            self.tabWidget.setCurrentIndex(self.get_tab_index_from_widget(widget))
            return

        technical_committee_widget = TechnicalCommitteeWidget(
            self.application, self.mutex, self.tabWidget
        )

        # create tab
        self.tabWidget.addTab(
            technical_committee_widget,
            self._("Technical Committee"),
        )

        self.repository_data_updated.connect(
            technical_committee_widget._on_fetch_members_from_network_async_worker_finished
        )

        self.tabWidget.setCurrentIndex(self.tabWidget.count() - 1)

    def update_title(self):
        """
        Update window title with version and currency

        :return:
        """
        self.setWindowTitle(
            "Tikka {version} - {currency}".format(  # pylint: disable=consider-using-f-string
                version=__version__,
                currency=self.application.currencies.get_current().name,
            )
        )

    def open_transfer_window(self) -> None:
        """
        Open transfer window

        :return:
        """
        TransferWindow(self.application, self.mutex, None, None, self).exec_()

    def open_scan_qrcode_window(self) -> None:
        """
        Open scan qrcode window

        :return:
        """
        logging.debug("create instance of ScanQRCodeWindow")
        # ScanQRCodeWindow(self.application, self).exec_()
        window = ScanQRCodeOpenCVWindow(self.application, self)
        if window.address is not None:
            # display window
            window.exec_()

    def open_add_address_window(self) -> None:
        """
        Open add address window

        :return:
        """
        AddressAddWindow(self.application, self).exec_()

    def open_import_account_window(self) -> None:
        """
        Open import account window

        :return:
        """
        AccountImportWindow(self.application, self).exec_()

    def open_create_account_window(self) -> None:
        """
        Open create account window

        :return:
        """
        AccountCreateWindow(self.application, self).exec_()

    def open_vault_import_by_mnemonic_window(self) -> None:
        """
        Open import vault by mnemonic window

        :return:
        """
        VaultImportByMnemonicWindow(self.application, self).exec_()

    def open_v1_import_account_window(self) -> None:
        """
        Open V1 import account window

        :return:
        """
        V1AccountImportWindow(self.application, self.mutex, self).exec_()

    def open_v1_import_in_v2_wizard_window(self) -> None:
        """
        Open V1 import in V2 wizard window

        :return:
        """
        V1AccountImportWizardWindow(self.application, self.mutex, self).exec_()

    def open_v1_import_file_window(self) -> None:
        """
        Open V1 import file window

        :return:
        """
        V1FileImportWindow(self.application, self.mutex, self).exec_()

    def open_configuration_window(self) -> None:
        """
        Open configuration window

        :return:
        """
        ConfigurationWindow(self.application, self).exec_()

    def open_about_window(self) -> None:
        """
        Open about window

        :return:
        """
        AboutWindow(self).exec_()

    def open_welcome_window(self) -> None:
        """
        Open welcome window

        :return:
        """
        WelcomeWindow(self.application, self).exec_()

    def open_save_user_data_window(self) -> None:
        """
        Open file dialog to save user data on disk as json file

        :return:
        """
        default_dir = self.application.repository.preferences.get(
            SAVE_DATA_DEFAULT_DIR_PREFERENCES_KEY
        )
        if default_dir is not None:
            default_dir = str(Path(default_dir).expanduser().absolute())
        else:
            default_dir = ""

        result = QFileDialog.getSaveFileName(
            self, self._("Save user data"), default_dir, "JSON Files (*.json)"
        )
        if result[0] == "":
            return None

        self.application.save_data(result[0])

        # update default dir preference
        self.application.repository.preferences.set(
            SAVE_DATA_DEFAULT_DIR_PREFERENCES_KEY,
            str(Path(result[0]).expanduser().absolute().parent),
        )

        self.statusbar.showMessage(self._("User data saved."), 5000)

        return None

    def open_restore_user_data_window(self) -> None:
        """
        Open file dialog to restore user data from json file

        :return:
        """
        default_dir = self.application.repository.preferences.get(
            SAVE_DATA_DEFAULT_DIR_PREFERENCES_KEY
        )
        if default_dir is not None:
            default_dir = str(Path(default_dir).expanduser().absolute())
        else:
            default_dir = ""

        result = QFileDialog.getOpenFileName(
            self, self._("Restore user data"), default_dir, "JSON Files (*.json)"
        )
        if result[0] == "":
            return None

        # disconnect servers
        self.application.connections.disconnect_all()

        # load user data
        self.application.load_data(result[0])

        # refresh widgets
        self.repository_data_updated.emit()
        logging.debug("repository_data_updated. Emit signal to widgets.")

        # connect to servers
        self.connect_to_node_async_worker.start()
        self.connect_to_indexer_async_worker.start()

        # update default dir preference
        self.application.repository.preferences.set(
            SAVE_DATA_DEFAULT_DIR_PREFERENCES_KEY,
            str(Path(result[0]).expanduser().absolute().parent),
        )

        self.statusbar.showMessage(self._("User data restored."), 5000)

        return None

    def on_currency_event(self, event: CurrencyEvent):
        """
        When a currency event is triggered

        :return:
        """
        if event.type == CurrencyEvent.EVENT_TYPE_PRE_CHANGE:
            self.save_tabs()
        else:
            self.update_title()
            self.init_tabs()

    def on_account_tree_double_click(self, index: QModelIndex):
        """
        When a row is double-clicked in account tree view

        :param index: QModelIndex instance
        :return:
        """
        account = index.internalPointer().element
        if isinstance(account, Account):
            fetch_transfers_from_network = False
            if self.get_account_widget(account.address) is None:
                fetch_transfers_from_network = True
            widget = self.add_account_tab(index.internalPointer().element)
            if fetch_transfers_from_network is True:
                widget.fetch_transfers_from_network_async_qworker.start()

    def on_account_table_double_click(self, index: QModelIndex):
        """
        When a row is double-clicked in account table

        :param index: QModelIndex instance
        :return:
        """
        table_view_row = index.internalPointer()
        account = self.application.accounts.get_by_address(table_view_row.address)
        if account is not None:
            fetch_transfers_from_network = False
            if self.get_account_widget(account.address) is None:
                fetch_transfers_from_network = True
            widget = self.add_account_tab(account)
            if fetch_transfers_from_network is True:
                widget.fetch_transfers_from_network_async_qworker.start()

    def on_add_account_event(self, event: AccountEvent) -> None:
        """
        Triggered when an account is created

        :param event: AccountEvent instance
        :return:
        """
        widget = self.add_account_tab(event.account)
        widget.fetch_transfers_from_network_async_qworker.start()

    def on_update_account_event(self, event: AccountEvent) -> None:
        """
        Triggered when an account is updated

        :param event: AccountEvent instance
        :return:
        """
        for index in range(0, self.tabWidget.count()):
            widget = self.tabWidget.widget(index)
            if (
                isinstance(widget, AccountWidget)
                and widget.account.address == event.account.address
            ):
                self.set_account_tab_title(index, widget.account)
                self.tabWidget.setTabIcon(
                    index, self.get_account_tab_icon(widget.account)
                )

    def on_delete_account_event(self, event: AccountEvent) -> None:
        """
        Triggered when an account is deleted

        :param event: AccountEvent instance
        :return:
        """
        for widget in self.get_tab_widgets_by_class(AccountWidget):
            if (
                isinstance(widget, AccountWidget)
                and widget.account.address == event.account.address
            ):
                self.tabWidget.removeTab(self.tabWidget.indexOf(widget))

    def _on_unit_changed(self) -> None:
        """
        Triggered when unit_combo_box selection changed

        :return:
        """
        unit_key = self.unit_combo_box.currentData()

        self.application.repository.preferences.set(
            SELECTED_UNIT_PREFERENCES_KEY, unit_key
        )
        self.application.event_dispatcher.dispatch_event(
            UnitEvent(UnitEvent.EVENT_TYPE_CHANGED)
        )

    def _on_node_connected(self, _=None):
        """
        Triggered when node is connected

        :return:
        """
        self.node_connection_status_icon.setIcon(QIcon(ICON_NETWORK_CONNECTED))
        self.statusbar.showMessage(self._("Loading data from network..."), 3000)
        self.fetch_data_from_network_node_async_qworker.start()

    def _on_node_disconnected(self, _=None):
        """
        Triggered when node is disconnected

        :return:
        """
        self.node_connection_status_icon.setIcon(QIcon(ICON_NETWORK_DISCONNECTED))

    def _on_indexer_connected(self, _=None):
        """
        Triggered when indexer is connected

        :return:
        """
        self.indexer_connection_status_icon.setIcon(QIcon(ICON_NETWORK_CONNECTED))

    def _on_indexer_disconnected(self, _=None):
        """
        Triggered when indexer is disconnected

        :return:
        """
        self.indexer_connection_status_icon.setIcon(QIcon(ICON_NETWORK_DISCONNECTED))

    def get_tab_widgets_by_class(self, widget_class: Any) -> List[QWidget]:
        """
        Return a list of widget which are instance of widget_class

        :param widget_class: Widget class
        :return:
        """
        widgets = []
        for index in range(0, self.tabWidget.count()):
            widget = self.tabWidget.widget(index)
            if isinstance(widget, widget_class):
                widgets.append(widget)

        return widgets

    def get_tab_index_from_widget(self, widget: QWidget) -> Optional[int]:
        """
        Return tab index of widget, or None if no tab with this widget

        :param widget: QWidget inherited instance
        :return:
        """
        for index in range(0, self.tabWidget.count()):
            if widget == self.tabWidget.widget(index):
                return index

        return None

    def get_account_widget(self, address: str) -> Optional[AccountWidget]:
        """
        Return AccountWidget instance if it is opened

        :param address: Account address
        :return:
        """
        for widget in self.get_tab_widgets_by_class(AccountWidget):
            if isinstance(widget, AccountWidget) and widget.account.address == address:
                return widget
        return None

    def connect_to_node(self):

        # init network connections
        if self.application.config.get(Config.RANDOM_CONNECTION_AT_START_KEY) is True:
            self.application.nodes.set_current_url_randomly()
        else:
            current_node = self.application.nodes.get(
                self.application.nodes.get_current_url()
            )
            if current_node is not None:
                # connect to node
                self.application.connections.node.connect(current_node)

    def _on_finished_connect_to_node(self):
        """
        Triggered when connection to node is finished

        :return:
        """
        if self.application.connections.node.is_connected():
            self.application.event_dispatcher.dispatch_event(
                ConnectionsEvent(ConnectionsEvent.EVENT_TYPE_NODE_CONNECTED)
            )

    def connect_to_indexer(self):
        if self.application.config.get(Config.RANDOM_CONNECTION_AT_START_KEY) is True:
            self.application.indexers.set_current_url_randomly()
        else:
            current_indexer = self.application.indexers.get(
                self.application.indexers.get_current_url()
            )
            if current_indexer is not None:
                # connect to indexer
                self.application.connections.indexer.connect(current_indexer)

    def _on_finished_connect_to_indexer(self):
        """
        Triggered when connection to indexer is finished

        :return:
        """
        if self.application.connections.indexer.is_connected():
            self.application.event_dispatcher.dispatch_event(
                ConnectionsEvent(ConnectionsEvent.EVENT_TYPE_INDEXER_CONNECTED)
            )

    def fetch_data_from_network_node(self):
        """
        Fetch data from the network node

        :return:
        """
        self.application.nodes.network_fetch_current_node()
        self.application.currencies.network_update_properties()

        accounts = self.application.accounts.get_list()
        try:
            self.application.accounts.network_update_balances(accounts)
        except Exception as exception:
            logging.exception(exception)

        addresses = [account.address for account in accounts]
        try:
            self.application.identities.network_update_identities(addresses)
        except Exception as exception:
            logging.exception(exception)
        else:
            identity_indice = self.application.identities.list_indice()
            try:
                self.application.smiths.network_update_smiths(identity_indice)
            except Exception as exception:
                logging.exception(exception)
            else:
                self.application.authorities.network_get_all()

        self.application.technical_committee.network_update_members()
        self.application.technical_committee.network_update_proposals()

    def _on_finished_fetch_data_from_network_node(self):
        """
        Triggered when async request fetch_data_from_network_node is finished

        :return:
        """
        self.statusbar.showMessage(self._("Data updated from network node."), 3000)
        self.repository_data_updated.emit()
        logging.debug("Main window network fetch finished....emit signal")

    def fetch_data_from_network_indexer(self):
        """
        Fetch data from the network indexer

        :return:
        """
        self.application.indexers.network_fetch_current_indexer()

    def _on_finished_fetch_data_from_network_indexer(self):
        """
        Triggered when async request fetch_data_from_network_indexer is finished

        :return:
        """
        self.statusbar.showMessage(self._("Data updated from network indexer."), 3000)
        self.repository_data_updated.emit()
        logging.debug("Main window network fetch finished....emit signal")


if __name__ == "__main__":
    qapp = QApplication(sys.argv)
    application_ = Application(DATA_PATH)
    MainWindow(application_).show()
    sys.exit(qapp.exec_())



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import abc
from typing import List, Optional
from uuid import UUID

from tikka.domains.entities.category import Category


class CategoriesRepositoryInterface(abc.ABC):
    """
    CategoriesRepositoryInterface class
    """

    COLUMN_ID = "category_id"
    COLUMN_NAME = "category_name"
    COLUMN_EXPANDED = "category_expanded"
    COLUMN_PARENT_ID = "category_parent_id"

    @abc.abstractmethod
    def add(self, category: Category) -> None:
        """
        Add a new category in repository

        :param category: Category instance
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get(self, id: UUID) -> Optional[Category]:  # pylint: disable=redefined-builtin
        """
        Return Category instance from repository

        :param id: Category ID
        :return:
        """
        raise NotImplementedError

    def update(self, category: Category) -> None:
        """
        Update category in repository

        :param category: Category instance
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def update_parent_id(self, parent_id: UUID, new_parent_id: Optional[UUID]):
        """
        Update all category with parent_id to new_parent_id

        :param parent_id: Parent ID
        :param new_parent_id: New parent ID
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def delete(self, id: UUID) -> None:  # pylint: disable=redefined-builtin
        """
        Delete category in repository

        :param id: Category ID to delete
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def delete_all(self) -> None:
        """
        Delete all categories in repository

        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def list(self, parent_id: Optional[UUID]) -> List[Category]:
        """
        Return categories with parent_id from repository

        :param parent_id: Category parent ID
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def list_all(self) -> List[Category]:
        """
        Return all categories by ascending alphabetical order from repository

        :return:
        """
        raise NotImplementedError
